import pytest
import numpy as np
import pandas as pd

from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import ShuffleSplit, KFold

from mllabs._experimenter import Experimenter
from mllabs._trainer import Trainer


class BadProcessor:
    __name__ = 'BadProcessor'
    def __init__(self, **kwargs):
        pass
    def fit(self, X, y=None):
        raise ValueError("intentional error")
    def transform(self, X):
        pass


@pytest.fixture
def sample_data():
    np.random.seed(42)
    n = 100
    return pd.DataFrame({
        'f1': np.random.randn(n),
        'f2': np.random.randn(n),
        'f3': np.random.randn(n),
        'target': np.random.randint(0, 2, n),
    })


@pytest.fixture
def exp(tmp_path, sample_data):
    e = Experimenter(
        data=sample_data,
        path=tmp_path / 'exp',
        sp=ShuffleSplit(n_splits=2, test_size=0.2, random_state=42),
        sp_v=KFold(n_splits=3, shuffle=True, random_state=42),
    )
    e.set_grp('scale', role='stage', processor=StandardScaler,
              method='transform', edges={'X': [(None, ['f1', 'f2', 'f3'])]})
    e.set_node('scaler', grp='scale')
    e.set_grp('model', role='head', processor=DecisionTreeClassifier,
              method='predict',
              edges={'X': [('scaler', None)], 'y': [(None, 'target')]},
              params={'max_depth': 3, 'random_state': 42})
    e.set_node('dt', grp='model')
    e.build()
    e.exp()
    return e


class TestAddTrainer:
    def test_default_splitter_same(self, exp):
        trainer = exp.add_trainer('t1')
        assert trainer.splitter is exp.sp_v
        assert trainer.splitter_params == exp.splitter_params

    def test_add_trainer_skip(self, exp):
        t1 = exp.add_trainer('t1')
        t2 = exp.add_trainer('t1', exist='skip')
        assert t1 is t2

    def test_add_trainer_error(self, exp):
        exp.add_trainer('t1')
        with pytest.raises(RuntimeError):
            exp.add_trainer('t1', exist='error')

    def test_splitter_same_with_params_raises(self, exp):
        with pytest.raises(ValueError):
            exp.add_trainer('t1', splitter='same', splitter_params={'y': 'target'})

    def test_custom_splitter(self, exp):
        sp = KFold(n_splits=5)
        trainer = exp.add_trainer('t1', splitter=sp)
        assert trainer.splitter is sp

    def test_no_splitter(self, exp):
        trainer = exp.add_trainer('t1', splitter=None)
        assert trainer.splitter is None
        assert trainer.split_indices is None
        assert trainer.get_n_splits() == 1


class TestSelectHead:
    def test_select_head_collects_upstream(self, exp):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        assert 'dt' in trainer.selected_heads
        assert 'scaler' in trainer.selected_stages

    def test_select_head_multiple(self, exp):
        exp.set_grp('model2', role='head', processor=DecisionTreeClassifier,
                    method='predict',
                    edges={'X': [('scaler', None)], 'y': [(None, 'target')]},
                    params={'max_depth': 5, 'random_state': 0})
        exp.set_node('dt2', grp='model2')
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        trainer.select_head(['dt2'])
        assert 'dt' in trainer.selected_heads
        assert 'dt2' in trainer.selected_heads
        assert 'scaler' in trainer.selected_stages


class TestTrain:
    def test_train_basic(self, exp):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        trainer.train()
        assert 'scaler' in trainer.node_objs
        assert trainer.node_objs['scaler'].status == 'built'
        assert 'dt' in trainer.node_objs
        assert trainer.node_objs['dt'].status == 'built'

    def test_train_skips_built(self, exp):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        trainer.train()
        objs_before = dict(trainer.node_objs)
        trainer.train()
        assert trainer.node_objs['scaler'] is objs_before['scaler']
        assert trainer.node_objs['dt'] is objs_before['dt']

    def test_train_no_splitter(self, exp):
        trainer = exp.add_trainer('t_nosplit', splitter=None)
        trainer.select_head(['dt'])
        trainer.train()
        assert trainer.node_objs['scaler'].status == 'built'
        assert trainer.node_objs['dt'].status == 'built'

    def test_train_error(self, exp):
        exp.set_grp('bad', role='head', processor=BadProcessor,
                    method='transform',
                    edges={'X': [(None, ['f1'])]})
        exp.set_node('bad_node', grp='bad')
        trainer = exp.add_trainer('t_err')
        trainer.select_head(['bad_node'])
        trainer.train()
        assert trainer.node_objs['bad_node'].status == 'error'
        err = trainer.node_objs['bad_node'].error
        assert err['type'] == 'ValueError'
        assert 'intentional error' in err['message']

    def test_train_error_continues_other_nodes(self, exp):
        exp.set_grp('bad', role='head', processor=BadProcessor,
                    method='transform',
                    edges={'X': [(None, ['f1'])]})
        exp.set_node('bad_node', grp='bad')
        trainer = exp.add_trainer('t_mixed')
        trainer.select_head(['dt', 'bad_node'])
        trainer.train()
        assert trainer.node_objs['dt'].status == 'built'
        assert trainer.node_objs['bad_node'].status == 'error'

    def test_train_n_splits(self, exp):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        assert trainer.get_n_splits() == 3
        trainer.train()
        objs = list(trainer.node_objs['dt'].get_obj())
        assert len(objs) == 3


class TestProcess:
    def test_process_yields_per_split(self, exp, sample_data):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        trainer.train()
        results = list(trainer.process(sample_data))
        assert len(results) == trainer.get_n_splits()

    def test_process_output_shape(self, exp, sample_data):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        trainer.train()
        for output in trainer.process(sample_data):
            assert output.get_shape()[0] == len(sample_data)


class TestResetNodes:
    def test_reset_clears_node_objs(self, exp):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        trainer.train()
        trainer.reset_nodes(['scaler'])
        assert 'scaler' not in trainer.node_objs
        assert 'dt' not in trainer.node_objs

    def test_reset_allows_retrain(self, exp):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        trainer.train()
        trainer.reset_nodes(['dt'])
        assert 'dt' not in trainer.node_objs
        trainer.train()
        assert trainer.node_objs['dt'].status == 'built'


class TestSaveLoad:
    def test_save_load_roundtrip(self, exp, sample_data):
        trainer = exp.add_trainer('t1')
        trainer.select_head(['dt'])
        trainer.train()
        path = exp.path

        loaded_exp = Experimenter.load(path, sample_data)
        loaded_trainer = loaded_exp.trainers['t1']
        assert loaded_trainer.name == 't1'
        assert set(loaded_trainer.selected_stages) == set(trainer.selected_stages)
        assert set(loaded_trainer.selected_heads) == set(trainer.selected_heads)
        assert 'scaler' in loaded_trainer.node_objs
        assert 'dt' in loaded_trainer.node_objs

    def test_save_creates_file(self, exp):
        trainer = exp.add_trainer('t1')
        assert (trainer.path / '__trainer.pkl').exists()



import re
import pickle
import pandas as pd
from ._node_processor import resolve_columns

class Metric:
    def __init__(
        self, name, experimenter, target_vars, output_var, metric_func, include_train = False
    ):
        self.experimenter = experimenter
        self.name = name
        self.target_vars = target_vars
        self.output_var = output_var
        self.include_train = include_train
        self.metric_func = metric_func
        self.metrics = dict()

    @property
    def metric_dir(self):
        """metric 데이터를 저장할 디렉토리 경로"""
        return self.experimenter.path / "__metric"

    @property
    def metric_path(self):
        """이 metric의 저장 파일 경로"""
        return self.metric_dir / f"{self.name}.pkl"

    def _ensure_metric_dir(self):
        """metric 디렉토리가 존재하는지 확인하고 없으면 생성"""
        if not self.metric_dir.exists():
            self.metric_dir.mkdir(parents=True, exist_ok=True)
        return self.metric_dir

    def save(self):
        """name, target_vars, output_var, include_train, metric_func, metrics를 파일로 저장"""
        self._ensure_metric_dir()
        data = {
            'name': self.name,
            'target_vars': self.target_vars,
            'output_var': self.output_var,
            'include_train': self.include_train,
            'metric_func': self.metric_func,
            'metrics': self.metrics
        }
        with open(self.metric_path, 'wb') as f:
            pickle.dump(data, f)

    def load(self):
        """파일에서 name, target_vars, output_var, include_train, metric_func, metrics를 로드"""
        if not self.metric_path.exists():
            raise FileNotFoundError(f"Metric data not found: {self.metric_path}")
        with open(self.metric_path, 'rb') as f:
            data = pickle.load(f)
        self.name = data['name']
        self.target_vars = data['target_vars']
        self.output_var = data['output_var']
        self.include_train = data['include_train']
        self.metric_func = data['metric_func']
        self.metrics = data['metrics']

    @classmethod
    def load_from_file(cls, experimenter, name):
        """저장된 Metric 정보를 불러와서 Metric 인스턴스 생성

        Args:
            experimenter: Experimenter 인스턴스
            name: metric 이름

        Returns:
            Metric: 복원된 Metric 인스턴스
        """
        filepath = experimenter.path / "__metric" / f"{name}.pkl"
        if not filepath.exists():
            raise FileNotFoundError(f"Metric data not found: {filepath}")

        with open(filepath, 'rb') as f:
            data = pickle.load(f)

        # Metric 인스턴스 생성
        metric = cls(
            name=data['name'],
            experimenter=experimenter,
            target_vars=data['target_vars'],
            output_var=data['output_var'],
            metric_func=data['metric_func'],
            include_train=data['include_train']
        )
        metric.metrics = data['metrics']

        return metric

    def _get_data(self, idx):
        # target_vars를 임시 key로 감싸서 get_data 호출
        temp_edges = {'_target': self.target_vars}
        result = list(self.experimenter.get_data(idx, temp_edges))
        # '_target' key의 데이터만 추출
        return [data_dict['_target'] for data_dict in result]

    def _get_metric(self, target_data, result_data):
        (true_train_t, true_train_v), true_valid = target_data
        selected_cols = resolve_columns(result_data['output_valid'], self.output_var)
        if len(selected_cols) == 0:
            return None
        prd = result_data['output_valid'].select_columns(selected_cols)
        result = {
            'valid': self.metric_func(true_valid.data, prd.data)
        }
        if self.include_train:
            prd = result_data['output_train'][0].select_columns(selected_cols)
            result['train_sub'] = self.metric_func(true_train_t.data, prd.data)
            if true_train_v is not None:
                prd = result_data['output_train'][1].select_columns(selected_cols)
                result['valid_sub'] = self.metric_func(true_train_v.data, prd.data)
        return result

    def _start(self, node):
        self.metrics[node] = list()

    def _set_metric(self, node, idx, metric):
        l = self.metrics[node]
        if len(l) != idx:
            raise RuntimeError("")
        l.append(metric)

    def _end(self, node):
        if len(self.metrics[node]) == 0:
            del self.metrics[node]
        self.save()

    def _get_nodes(self, nodes):
        if nodes is None:
            # 기존 동작: 모든 노드
            node_names = list(self.metrics.keys())
        elif isinstance(nodes, list):
            node_names = [n for n in nodes if n in self.metrics]
        elif isinstance(nodes, str):
            pat = re.compile(nodes)
            node_names = [k for k in self.metrics.keys() if k is not None and pat.search(k)]
        else:
            raise ValueError(f"nodes must be None, list, or str, got {type(nodes)}")
        return node_names
    
    def reset_nodes(self, nodes):
        for node in nodes:
            if node not in self.metrics:
                continue
            del self.metrics[node]
        self.save()

    def get_metric(self, node):
        l = list()
        for i, sub in enumerate(self.metrics[node]):
            l.append(
                pd.concat([pd.Series(j, name = str(no)) for no, j in enumerate(sub)], axis=1).unstack()
            )
        return pd.concat(l, axis=1).unstack(level=[0, 1]).rename(node)

    def get_metrics(self, nodes):
        node_names = self._get_nodes(nodes)
        return pd.concat([self.get_metric(node) for node in node_names], axis=1).T

    def get_metrics_agg(self, nodes, inner_fold = True, outer_fold = True, include_std = False):
        if outer_fold and not inner_fold:
            raise ValueError("")
        df = self.get_metrics(nodes)
        if inner_fold:
            df_agg_mean = df.stack(level = 1).groupby(level=0).mean()
            if include_std:
                df_agg_std = df.stack(level = 1).groupby(level=0).std()
            else:
                df_agg_std = None
            if outer_fold:
                df_agg_mean = df_agg_mean.stack(level=0).groupby(level=0).mean()
                if include_std:
                    df_agg_std = df_agg_std.stack(level = 0).groupby(level=0).mean()
            return df_agg_mean, df_agg_std
        return df


import pickle
import shutil

from ._base import Collector
from .._node_processor import resolve_columns


class OutputCollector(Collector):
    def __init__(self, name, connector, output_var, include_target=True):
        super().__init__(name, connector)
        self.output_var = output_var
        self.include_target = include_target

    def _collect(self, node, idx, inner_idx, context):
        cols = resolve_columns(context['output_valid'], self.output_var)
        if len(cols) == 0:
            return

        output_valid = context['output_valid'].select_columns(cols)
        train_sub = context['output_train'][0].select_columns(cols)
        valid_sub = context['output_train'][1]
        if valid_sub is not None:
            valid_sub = valid_sub.select_columns(cols)

        entry = {
            'output_train': (
                train_sub.to_array(),
                valid_sub.to_array() if valid_sub is not None else None
            ),
            'output_valid': output_valid.to_array(),
            'columns': output_valid.get_columns(),
        }

        self._ensure_path()
        node_dir = self.path / node
        node_dir.mkdir(parents=True, exist_ok=True)
        with open(node_dir / f"{idx}_{inner_idx}.pkl", 'wb') as f:
            pickle.dump(entry, f)

    def has_node(self, node):
        if self.path is not None:
            node_dir = self.path / node
            return node_dir.exists() and any(node_dir.glob("*.pkl"))
        return False

    def reset_nodes(self, nodes):
        for node in nodes:
            if self.path is not None:
                node_dir = self.path / node
                if node_dir.exists():
                    shutil.rmtree(node_dir)

    def save(self):
        if self.path is None:
            return
        self._ensure_path()
        config = {
            'name': self.name,
            'connector': self.connector,
            'output_var': self.output_var,
            'include_target': self.include_target,
        }
        with open(self.path / '__config.pkl', 'wb') as f:
            pickle.dump(config, f)

    @classmethod
    def load(cls, path):
        with open(path / '__config.pkl', 'rb') as f:
            config = pickle.load(f)
        obj = cls(
            name=config['name'],
            connector=config['connector'],
            output_var=config['output_var'],
            include_target=config['include_target'],
        )
        obj.path = path
        return obj

    def _get_saved_nodes(self):
        if self.path is None or not self.path.exists():
            return []
        return [d.name for d in self.path.iterdir() if d.is_dir()]

    def get_output(self, node, idx, inner_idx):
        file_path = self.path / node / f"{idx}_{inner_idx}.pkl"
        if not file_path.exists():
            raise FileNotFoundError(f"Output data not found: {file_path}")
        with open(file_path, 'rb') as f:
            return pickle.load(f)

    def get_outputs(self, node):
        node_dir = self.path / node
        if not node_dir.exists():
            raise FileNotFoundError(f"Node directory not found: {node_dir}")
        results = {}
        for f in sorted(node_dir.glob("*.pkl")):
            parts = f.stem.split('_')
            idx, inner_idx = int(parts[0]), int(parts[1])
            with open(f, 'rb') as fp:
                results[(idx, inner_idx)] = pickle.load(fp)
        return results



from ._base import Collector
from ._metric import MetricCollector
from ._stacking import StackingCollector
from ._model_attr import ModelAttrCollector
from ._output import OutputCollector

try:
    from ._shap import SHAPCollector
except ImportError:
    SHAPCollector = None



import numpy as np

from ._base import DataFilter


class IndexFilter(DataFilter):
    def __init__(self, index):
        self.index = index

    def _select(self, data_dict):
        first_val = next(iter(data_dict.values()))
        data_index = first_val.get_index()
        mask = np.isin(data_index, self.index)
        return np.where(mask)[0]



import pandas as pd

def desc_spec(exp):
    """실험 스펙을 Markdown으로 반환"""
    lines = []

    # 실험 타이틀
    if exp.title:
        lines.append(f"## {exp.title}")
        lines.append("")

    lines.append("| 항목 | 값 |")
    lines.append("|------|-----|")

    # Outer Splitter (sp)
    sp_name = type(exp.sp).__name__
    sp_params = []
    if hasattr(exp.sp, 'n_splits'):
        sp_params.append(f"n_splits={exp.sp.n_splits}")
    if hasattr(exp.sp, 'random_state') and exp.sp.random_state is not None:
        sp_params.append(f"random_state={exp.sp.random_state}")
    if hasattr(exp.sp, 'test_size') and exp.sp.test_size is not None:
        sp_params.append(f"test_size={exp.sp.test_size}")
    if hasattr(exp.sp, 'shuffle'):
        sp_params.append(f"shuffle={exp.sp.shuffle}")
    sp_info = f"{sp_name}({', '.join(sp_params)})" if sp_params else sp_name
    lines.append(f"| **Outer Splitter (sp)** | `{sp_info}` |")

    # Inner Splitter (sp_v)
    if exp.sp_v is not None:
        sp_v_name = type(exp.sp_v).__name__
        sp_v_params = []
        if hasattr(exp.sp_v, 'n_splits'):
            sp_v_params.append(f"n_splits={exp.sp_v.n_splits}")
        if hasattr(exp.sp_v, 'random_state') and exp.sp_v.random_state is not None:
            sp_v_params.append(f"random_state={exp.sp_v.random_state}")
        if hasattr(exp.sp_v, 'test_size') and exp.sp_v.test_size is not None:
            sp_v_params.append(f"test_size={exp.sp_v.test_size}")
        if hasattr(exp.sp_v, 'shuffle'):
            sp_v_params.append(f"shuffle={exp.sp_v.shuffle}")
        sp_v_info = f"{sp_v_name}({', '.join(sp_v_params)})" if sp_v_params else sp_v_name
        lines.append(f"| **Inner Splitter (sp_v)** | `{sp_v_info}` |")
    else:
        lines.append(f"| **Inner Splitter (sp_v)** | None |")

    # Splitter Params
    if exp.splitter_params:
        params_str = ", ".join([f"{k}='{v}'" for k, v in exp.splitter_params.items()])
        lines.append(f"| **Splitter Params** | `{{{params_str}}}` |")
    else:
        lines.append(f"| **Splitter Params** | `{{}}` |")

    # Fold 수
    lines.append(f"| **Outer Folds** | {len(exp.train_idx_list)} |")
    if len(exp.train_idx_list) > 0:
        inner_folds = len(exp.train_idx_list[0])
        lines.append(f"| **Inner Folds** | {inner_folds} |")

    return "\n".join(lines)

def desc_pipeline(pipeline, max_depth=None, direction='TD'):
    """파이프라인 구조를 Mermaid Markdown으로 반환

    Args:
        pipeline: Pipeline 인스턴스
        max_depth: 최대 표시 깊이 (None이면 무제한)
        direction: 그래프 방향 ('TD': Top-Down, 'LR': Left-Right)
    """
    # 노드 개수 계산 함수
    def count_nodes_in_group(grp):
        count = len(grp.nodes)
        for child_grp_name in grp.children:
            child_grp = pipeline.grps[child_grp_name]
            count += count_nodes_in_group(child_grp)
        return count

    # 1. Node 단위 우선순위 생성 (BFS)
    node_priorities = {}
    queue = [('DataSource', 1)]

    while queue:
        current_node, priority = queue.pop(0)

        # 이미 더 낮은 우선순위(더 상위)가 할당되었으면 스킵
        if current_node in node_priorities:
            continue

        node_priorities[current_node] = priority

        # current_node를 edge로 가지는 child 노드들 찾기
        for name in pipeline.nodes.keys():
            if name is not None:
                node_attrs = pipeline.get_node_attrs(name)
                for key, edge_list in node_attrs['edges'].items():
                    for edge_name, _ in edge_list:
                        if (current_node == 'DataSource' and edge_name is None) or (edge_name == current_node):
                            # child 노드 발견
                            if name not in node_priorities:
                                queue.append((name, priority + 1))

    # 2. Group 단위 우선순위 생성 (포함된 노드 중 가장 낮은 우선순위 = 가장 상위)
    grp_priorities = {}
    for grp_name, grp in pipeline.grps.items():
        if len(grp.nodes) > 0:
            grp_priorities[grp_name] = min(node_priorities.get(node_name, float('inf')) for node_name in grp.nodes)
        else:
            grp_priorities[grp_name] = float('inf')

    # 3. 소속 그룹이 없는 노드와 최상위 그룹 수집
    grouped_nodes = set()
    for grp in pipeline.grps.values():
        grouped_nodes.update(grp.nodes)

    ungrouped_nodes = [name for name in pipeline.nodes.keys() if name is not None and name not in grouped_nodes]
    top_level_items = []

    for grp_name, grp in pipeline.grps.items():
        if grp.parent is None:
            top_level_items.append(('group', grp))

    for node_name in ungrouped_nodes:
        if node_name in pipeline.nodes:
            top_level_items.append(('node', pipeline.nodes[node_name]))

    # 우선순위로 정렬
    def get_priority(item):
        item_type, obj = item
        if item_type == 'group':
            return grp_priorities.get(obj.name, float('inf'))
        else:
            return node_priorities.get(obj.name, float('inf'))

    top_level_items.sort(key=get_priority)

    # 4. Mermaid 생성
    lines = []
    lines.append("```mermaid")
    lines.append(f"graph {direction}")
    lines.append("")

    # DataSource 노드
    lines.append("    DataSource([DataSource])")
    lines.append("    style DataSource fill:#fff9c4,stroke:#f57c00,stroke-width:3px")
    lines.append("")

    # Recursive 함수로 그룹과 노드 생성
    def render_group(grp, indent=4, current_depth=1):
        indent_str = " " * indent
        result = []
        result.append(f"{indent_str}subgraph grp_{grp.name}[\"{grp.name}\"]")

        # max_depth에 도달했으면 노드 개수만 표시
        if max_depth is not None and current_depth >= max_depth:
            node_count = count_nodes_in_group(grp)
            result.append(f"{indent_str}    grp_{grp.name}_count[\"{node_count} node(s)\"]")
            result.append(f"{indent_str}    style grp_{grp.name}_count fill:#f5f5f5,stroke:#9e9e9e,stroke-dasharray: 5 5")
        else:
            items = []
            for child_grp_name in grp.children:
                child_grp = pipeline.grps[child_grp_name]
                items.append(('group', child_grp))
            for node_name in grp.nodes:
                if node_name in pipeline.nodes:
                    items.append(('node', pipeline.nodes[node_name]))

            # 우선순위로 정렬
            items.sort(key=get_priority)

            # 렌더링
            for item_type, obj in items:
                if item_type == 'group':
                    result.extend(render_group(obj, indent + 4, current_depth + 1))
                else:
                    node_name = obj.name
                    result.append(f"{indent_str}    node_{node_name}[\"{node_name}\"]")
                    result.append(f"{indent_str}    style node_{node_name} fill:#c8e6c9,stroke:#388e3c,stroke-width:2px")

        result.append(f"{indent_str}end")
        result.append(f"{indent_str}style grp_{grp.name} fill:#e3f2fd,stroke:#1976d2,stroke-width:2px")
        return result

    # Top-level items 렌더링
    for item_type, obj in top_level_items:
        if item_type == 'group':
            # top-level 그룹은 depth 1부터 시작
            if max_depth is None or max_depth >= 1:
                lines.extend(render_group(obj, indent=4, current_depth=1))
                lines.append("")
        else:
            # top-level 노드도 depth 1
            if max_depth is None or max_depth >= 1:
                node_name = obj.name
                lines.append(f"    node_{node_name}[\"{node_name}\"]")
                lines.append(f"    style node_{node_name} fill:#c8e6c9,stroke:#388e3c,stroke-width:2px")
                lines.append("")

    # Edge 연결 알고리즘
    # 1. 각 노드가 속한 최상위 노드를 담는 딕셔너리
    node_to_top = {}
    for item_type, obj in top_level_items:
        if item_type == 'group':
            # 그룹에 속한 모든 노드 수집 (recursive)
            def collect_nodes_in_group(grp):
                nodes = []
                for node_name in grp.nodes:
                    nodes.append(node_name)
                for child_grp_name in grp.children:
                    child_grp = pipeline.grps[child_grp_name]
                    nodes.extend(collect_nodes_in_group(child_grp))
                return nodes

            nodes_in_grp = collect_nodes_in_group(obj)
            for node_name in nodes_in_grp:
                node_to_top[node_name] = ('group', obj.name)
        else:
            node_to_top[obj.name] = ('node', obj.name)

    # 2. 상위 노드에서 하위 노드를 DFS 탐색하며 incoming 연결 수집
    top_node_incoming = {}  # top_node -> set of incoming top_nodes

    def dfs_collect_incoming(top_item_type, top_item_name):
        incoming = set()

        # top_item에 속한 모든 노드 찾기
        if top_item_type == 'group':
            grp = pipeline.grps[top_item_name]
            def collect_nodes_in_group(grp):
                nodes = []
                for node_name in grp.nodes:
                    nodes.append(node_name)
                for child_grp_name in grp.children:
                    child_grp = pipeline.grps[child_grp_name]
                    nodes.extend(collect_nodes_in_group(child_grp))
                return nodes
            nodes = collect_nodes_in_group(grp)
        else:
            nodes = [top_item_name]

        # 각 노드의 edges 확인
        for node_name in nodes:
            if node_name in pipeline.nodes:
                node_attrs = pipeline.get_node_attrs(node_name)
                for key, edge_list in node_attrs['edges'].items():
                    for edge_name, edge_var in edge_list:
                        if edge_name is None:
                            # DataSource 연결
                            incoming.add(('datasource', 'DataSource'))
                        elif edge_name in node_to_top:
                            # edge 노드의 최상위 노드 찾기
                            edge_top = node_to_top[edge_name]
                            # 같은 top node 내부 연결은 제외
                            if not (top_item_type == edge_top[0] and top_item_name == edge_top[1]):
                                incoming.add(edge_top)

        return incoming

    # 각 top-level item의 incoming 수집
    for item_type, obj in top_level_items:
        if item_type == 'group':
            key = ('group', obj.name)
        else:
            key = ('node', obj.name)
        top_node_incoming[key] = dfs_collect_incoming(item_type, obj.name if item_type == 'group' else obj.name)

    # 3. Edge 출력
    edges_set = set()
    for target, incoming_set in top_node_incoming.items():
        target_type, target_name = target
        target_id = f"grp_{target_name}" if target_type == 'group' else f"node_{target_name}"

        for source_type, source_name in incoming_set:
            if source_type == 'datasource':
                source_id = "DataSource"
            elif source_type == 'group':
                source_id = f"grp_{source_name}"
            else:
                source_id = f"node_{source_name}"

            edges_set.add((source_id, target_id))

    for source, target in sorted(edges_set):
        lines.append(f"    {source} --> {target}")

    lines.append("```")

    return "\n".join(lines)

def desc_node(pipeline, node_name, direction='TD', show_params=False):
    """특정 노드까지의 연결 구조를 Mermaid Markdown으로 반환

    Args:
        pipeline: Pipeline 인스턴스
        node_name: 대상 노드 이름
        direction: 그래프 방향 ('TD': Top-Down, 'LR': Left-Right)
        show_params: True이면 노드의 파라미터 정보를 표시 (default: False)
    """
    if node_name not in pipeline.nodes or node_name is None:
        return f"Node '{node_name}' not found"

    # DataSource에서 node_name까지의 경로 찾기 (BFS)
    def find_paths_to_node(target):
        paths = []
        queue = [(['DataSource'], set(['DataSource']))]

        while queue:
            path, visited = queue.pop(0)
            current = path[-1]

            # target에 도달했으면 경로 저장
            if current == target:
                paths.append(path[:])
                continue

            # current를 edge로 가지는 노드들 찾기
            for name in pipeline.nodes.keys():
                if name is not None and name not in visited:
                    found = False
                    node_attrs = pipeline.get_node_attrs(name)
                    edges = node_attrs['edges']
                    for key, edge_list in edges.items():
                        if found:
                            break
                        for edge_name, _ in edge_list:
                            if (current == 'DataSource' and edge_name is None) or (edge_name == current):
                                new_path = path + [name]
                                new_visited = visited | {name}
                                queue.append((new_path, new_visited))
                                found = True
                                break
        return paths

    paths = find_paths_to_node(node_name)

    if not paths:
        return f"No path from DataSource to '{node_name}'"

    # Mermaid 생성
    lines = []
    lines.append("```mermaid")
    lines.append(f"graph {direction}")
    lines.append("")

    # DataSource 노드
    lines.append("    DataSource([DataSource])")
    lines.append("    style DataSource fill:#fff9c4,stroke:#f57c00,stroke-width:3px")
    lines.append("")

    # 경로에 포함된 모든 노드 수집
    all_nodes = set()
    for path in paths:
        all_nodes.update(path)
    all_nodes.discard('DataSource')

    # 노드의 grp 경로를 구하는 헬퍼
    def get_grp_path(node_name):
        if node_name is None:
            return node_name
        parts = []
        node_obj = pipeline.get_node(node_name)
        if node_obj is None:
            return node_name
        grp_obj = pipeline.get_grp(node_obj.grp)
        while grp_obj is not None:
            parts.insert(0, grp_obj.name)
            grp_obj = pipeline.get_grp(grp_obj.parent)
        parts.append(node_name)
        return '/'.join(parts)

    # 각 노드를 subgraph로 생성
    for name in sorted(all_nodes):
        if name in pipeline.nodes:
            node_attrs = pipeline.get_node_attrs(name)
            edges = node_attrs['edges']
            display_name = get_grp_path(name)
            lines.append(f"    subgraph node_{name}[\"{display_name}\"]")

            if show_params:
                # 파라미터 정보 포맷팅
                processor_name = node_attrs['processor'].__name__ if node_attrs['processor'] else 'None'
                method = node_attrs['method']

                info_parts = ["<table>"]
                info_parts.append(f"<tr><td align='left'><b>processor</b></td><td align='left'>{processor_name}</td></tr>")
                info_parts.append(f"<tr><td align='left'><b>method</b></td><td align='left'>{method}</td></tr>")

                # params 정보
                if node_attrs['params']:
                    for key, value in node_attrs['params'].items():
                        value_str = str(value)
                        if len(value_str) > 40:
                            value_str = value_str[:37] + '...'
                        info_parts.append(f"<tr><td align='left'><b>{key}</b></td><td align='left'>{value_str}</td></tr>")
                info_parts.append("</table>")
                params_content = "".join(info_parts)
                lines.append(f"        {name}_info[\"{params_content}\"]")
            else:
                # show_params가 False면 빈 더미 노드
                lines.append(f"        {name}_dummy[ ]")
                lines.append(f"        style {name}_dummy fill:none,stroke:none")

            lines.append(f"    end")

            # target 노드는 다른 색으로 표시
            if name == node_name:
                lines.append(f"    style node_{name} fill:#ffcdd2,stroke:#c62828,stroke-width:3px")
            else:
                lines.append(f"    style node_{name} fill:#c8e6c9,stroke:#388e3c,stroke-width:2px")
            lines.append("")

    # 경로상의 엣지 수집 (key별로 구분)
    # edges_dict: {(source, target): set of keys}
    edges_dict = {}
    for name in all_nodes:
        if name in pipeline.nodes:
            node_attrs = pipeline.get_node_attrs(name)
            edges = node_attrs['edges']
            for key, edge_list in edges.items():
                for edge_name, _ in edge_list:
                    if edge_name is None:
                        source = "DataSource"
                    else:
                        source = f"node_{edge_name}"
                    target = f"node_{name}"
                    # source가 경로에 포함된 경우만
                    source_node = edge_name if edge_name else 'DataSource'
                    if source_node in all_nodes or source_node == 'DataSource':
                        edge_key = (source, target)
                        if edge_key not in edges_dict:
                            edges_dict[edge_key] = set()
                        edges_dict[edge_key].add(key)

    # 엣지 출력 (key 표시)
    for (source, target), keys in sorted(edges_dict.items()):
        keys_str = ','.join(sorted(keys))
        if keys_str != 'X':
            lines.append(f"    {source} -->|{keys_str}| {target}")
        else:
            lines.append(f"    {source} --> {target}")

    lines.append("```")
    lines.append("")
    target_display = get_grp_path(node_name)
    lines.append(f"**Path from DataSource to '{target_display}' ({len(paths)} path(s) found)**")

    # Edge 정보 테이블 추가
    node_attrs = pipeline.get_node_attrs(node_name)
    edges = node_attrs['edges']
    lines.append("")
    lines.append("### Edges")
    lines.append("")
    lines.append("| Key | Node | Var |")
    lines.append("|-----|------|-----|")
    for key in sorted(edges.keys()):
        edge_list = edges[key]
        for edge_name, var_spec in edge_list:
            if edge_name is None:
                node_display = "Data Source"
            else:
                if edge_name:
                    node_display = get_grp_path(edge_name)
                else:
                    node_display = edge_name
            var_display = "*" if var_spec is None else f"`{var_spec}`"
            lines.append(f"| {key} | {node_display} | {var_display} |")

    return "\n".join(lines)

def desc_status(exp):
    stage_nodes = []
    head_nodes = []
    for name in exp.pipeline.nodes.keys():
        if name is None:
            continue
        node = exp.pipeline.get_node(name)
        grp = exp.pipeline.get_grp(node.grp)
        if grp.role == 'stage':
            stage_nodes.append(name)
        elif grp.role == 'head':
            head_nodes.append(name)

    def _get_status(name):
        if name not in exp.node_objs:
            return 'init'
        return exp.node_objs[name].status

    def _status_summary(nodes):
        counts = {}
        for name in nodes:
            s = _get_status(name)
            counts[s] = counts.get(s, 0) + 1
        return counts

    lines = []

    # Experiment status
    exp_status = getattr(exp, 'status', 'open')
    lines.append(f"**Experiment**: {exp_status}")
    lines.append("")

    # Stage Nodes
    stage_counts = _status_summary(stage_nodes)
    lines.append(f"**Stage Nodes** ({len(stage_nodes)})")
    lines.append("")
    if stage_nodes:
        parts = [f"{s}: {c}" for s, c in sorted(stage_counts.items())]
        lines.append(f"| {' | '.join(stage_counts.keys())} |")
        lines.append(f"| {' | '.join(['---'] * len(stage_counts))} |")
        lines.append(f"| {' | '.join(str(c) for c in stage_counts.values())} |")
    lines.append("")

    # Head Nodes
    head_counts = _status_summary(head_nodes)
    lines.append(f"**Head Nodes** ({len(head_nodes)})")
    lines.append("")
    if head_nodes:
        lines.append(f"| {' | '.join(head_counts.keys())} |")
        lines.append(f"| {' | '.join(['---'] * len(head_counts))} |")
        lines.append(f"| {' | '.join(str(c) for c in head_counts.values())} |")
    lines.append("")

    # Error details
    error_nodes = []
    for name in stage_nodes + head_nodes:
        if name in exp.node_objs and exp.node_objs[name].status == 'error':
            error_nodes.append(name)

    if error_nodes:
        lines.append(f"**Errors** ({len(error_nodes)})")
        lines.append("")
        for name in error_nodes:
            err = exp.node_objs[name].error
            node = exp.pipeline.get_node(name)
            grp = exp.pipeline.get_grp(node.grp)
            lines.append(f"### {name} ({grp.role})")
            lines.append(f"- **fold**: {err['fold']}")
            lines.append(f"- **{err['type']}**: {err['message']}")
            lines.append(f"```\n{err['traceback']}```")
            lines.append("")

    return "\n".join(lines)


def desc_obj_vars(exp, obj_vars):
    # 첫 번째 항목 사용 (가장 빈도 높은 것)
    input_vars, output_vars, fold_indices = obj_vars

    # 입력 변수 DataFrame 생성
    input_data = []
    for var in input_vars:
        if '__' in var:
            node = var.split('__')[0]
        else:
            node = 'DataSource'
        input_data.append({'node': node, 'name': var})

    if input_data:
        input_df = pd.DataFrame(input_data)
        # 노드별로 일련번호 부여
        input_df['seq'] = input_df.groupby('node').cumcount()
        input_df = input_df.set_index(['node', 'seq'])[['name']]
    else:
        input_df = pd.DataFrame(columns=['name'])

    # 출력 변수 DataFrame 생성
    if output_vars:
        output_df = pd.DataFrame({'name': output_vars})
    else:
        output_df = pd.DataFrame(columns=['name'])

    return input_df, output_df



import os
import pandas as pd
import numpy as np
import polars as pl
import polars.selectors as cs
from functools import partial
from itertools import product
from collections import OrderedDict

def get_type_df(df):
    """
    get data type of polars df
    Parameters:
        df: pl.DataFrame, pl.LazyFrame
            DataFrame to examine the proper data type
    Returns:
        pd.DataFrame
            DataFrame which contains the information about the types
    Examples:
        >>> df_type = pl.scan_csv('test.csv').pipe(get_type_df)
    """
    stat = [
        ('min', cs.numeric().min()),
        ('max', cs.numeric().max()),
        ('na', cs.all().is_null().sum()),
        ('count', cs.all().count()),
        ('n_unique', cs.all().n_unique()),
    ]
    df_type = df.select(
        **{k: pl.struct(v) for k, v in stat}
    ).melt().unnest('value').rename({'variable': 'stat'})\
    .melt(id_vars='stat', variable_name='feature')
    if type(df_type) == pl.LazyFrame:
        df_type = df_type.collect()
    if type(df) == pl.lazyframe.frame.LazyFrame:
        dtypes = df.collect_schema().dtypes()
        columns = df.collect_schema().names()
    else:
        dtypes = df.dtypes
        columns = df.columns
    df_type = df_type.pivot(
        index='feature', columns='stat', values='value'
    ).to_pandas().assign(
        feature = lambda x: x['feature'].astype(str)
    ).set_index('feature').join(
        pd.Series([str(i) for i in dtypes], index=columns, name='dtype')
    )
    for i, mn, mx in [
        ('f32', np.finfo(np.float32).min, np.finfo(np.float32).max),
        ('i32', np.iinfo(np.int32).min, np.iinfo(np.int32).max),
        ('i16', np.iinfo(np.int16).min, np.iinfo(np.int16).max),
        ('i8', np.iinfo(np.int8).min, np.iinfo(np.int8).max)
    ]:
        df_type[i] = df_type.loc[df_type['dtype'] != 'String'].apply(
            lambda x: (x['min'] >= mn) and (x['max'] <= mx), axis=1
        )
    return df_type

def get_type_vars(var_list):
    """
    Processes a list of variables and generates a DataFrame summarizing their properties, such as 
    the number of unique values, count, and range (min, max). Also checks whether numeric variables 
    fit within certain data type ranges (e.g., float32, int32, int16, int8).

    Args:
        var_list (list of tuples): A list of tuples where each tuple contains the following elements:
            - src (str): The source or origin of the variable.
            - s (pd.Series): The pandas Series representing the variable.
            - desc (str): A description of the variable (unused in the function).

    Returns:
        pd.DataFrame: A pandas DataFrame indexed by variable name with the following columns:
            - 'src' (str): The source of the variable.
            - 'dtype' (str): The data type of the variable.
            - 'na' (int): The number of missing values in the variable.
            - 'n_unique' (int): The number of unique values in the variable.
            - 'count' (int): The total count of non-missing values.
            - 'min' (varied): The minimum value of the variable (if applicable).
            - 'max' (varied): The maximum value of the variable (if applicable).
            - 'f32' (bool): Whether the variable can fit within a float32 data type range.
            - 'i32' (bool): Whether the variable can fit within an int32 data type range.
            - 'i16' (bool): Whether the variable can fit within an int16 data type range.
            - 'i8' (bool): Whether the variable can fit within an int8 data type range.

    Example:
        >>> var_list = [
        ...     ('source1', pd.Series([1, 2, 3], dtype='int32'), 'description1', 'Numeric'),
        ...     ('source2', pd.Series(['a', 'b', 'a'], dtype='category'), 'description2', 'Categorical')
        ... ]
        >>> get_type_vars(var_list)
        # Returns a DataFrame with statistics on the variables, including dtype compatibility.
    """
    s_list = list()
    for src, s, desc in var_list:
        if str(s.dtype) == 'category':
            t = 'Categorical'
            if s.dtype.ordered:
                s_list.append(
                    pd.Series(
                        [src, s.name, desc, t, s.isna().sum()] + s.agg(['nunique', 'count', 'min', 'max']).tolist(),
                        index=['src', 'var', 'Description', 'dtype', 'na', 'n_unique', 'count', 'min', 'max']
                    )
                )
            else:
                s_list.append(
                    pd.Series(
                        [src, s.name, desc, t, s.isna().sum()] + s.agg(['nunique', 'count']).tolist(),
                        index=['src', 'var', 'Description', 'dtype', 'na', 'n_unique', 'count']
                    )
                )
        else:
            t = str(s.dtype)
            if t == 'str':
                t = 'String'
            if t == 'datetime64[ns]':
                t = 'Datetime'
            else:
                t = t[:1].upper() + t[1:]
            s_list.append(
                pd.Series(
                    [src, s.name, desc, s.isna().sum(), t] + s.agg(['nunique', 'count', 'min', 'max']).tolist(),
                    index=['src', 'var', 'Description', 'na', 'dtype', 'n_unique', 'count', 'min', 'max']
                )
            )
    df = pd.DataFrame(s_list).set_index('var')
    if 'min' in df.columns:
        for i, mn, mx in [
            ('f32', np.finfo(np.float32).min, np.finfo(np.float32).max),
            ('i32', np.iinfo(np.int32).min, np.iinfo(np.int32).max),
            ('i16', np.iinfo(np.int16).min, np.iinfo(np.int16).max),
            ('i8', np.iinfo(np.int8).min, np.iinfo(np.int8).max)
        ]:
            df[i] = df.loc[~df['dtype'].isin(['String', 'Categorical', 'Datetime'])].apply(
                lambda x: (x['min'] >= mn) and (x['max'] <= mx), axis=1
            )
    return df

def merge_type_df(dfs):
    """
    merge type data frame
    Parameters:
        dfs: list
            merge the type information of DataFrames
    Returns:
        df - return DataFrame
    Examples:
        >>> merge_type_df([pl.scan_csv(i).pipe(get_type_df) for i in glob(*.csv)])
    """
    return pd.concat(dfs, axis=0).pipe(lambda x: x.groupby(x.index).agg(
        {'min': 'min', 'max': 'max', 'na': 'sum', 'count': 'sum', 'n_unique': 'mean', 'dtype': 'max', 'f32': 'all', 'i32': 'all', 'i16': 'all', 'i8': 'all'})
    )

def get_type_pl(df_type, predefine={}, f32=True, i64=False, cat_max=np.inf, txt_cols = []):
    """
    get datatype for Polars DataFrame
    Parameters:
        df_type: pd.DataFrame
            the DataFrame whichi contains type information
        predefine: dict
            the predefined data type
        f32: Boolean
            Use f32 if possible
        i64: Boolean
            use i64 for all integer types
        cat_max: Int
            maximum number of categories for categorical types
        txt_cols: list
            Text type columns
    Returns:
        dict
            the dictionary of type mapping for pl.DataFrame loading
    Examples:
        >>> df_type = pl.scan_csv('test.csv').pipe(get_type_df)
        >>> dt = get_type_pl(df_type, {'id': pl.Int64})
    """
    ret_type = predefine.copy()
    
    for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Date')].index:
        ret_type = pl.Datetime
    if f32:
        for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Int') & (df_type['na'] > 0) & (df_type['f32'])].index:
            ret_type[i] = pl.Float32
        for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Float') & (df_type['f32'])].index:
            ret_type[i] = pl.Float32
    for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Int') & (df_type['na'] > 0)].index:
        ret_type[i] = pl.Float64
    for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Float')].index:
        ret_type[i] = pl.Float64
    if not i64:
        for col, ty in [('i8', pl.Int8), ('i16', pl.Int16), ('i32', pl.Int32)]:
            for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Int') & df_type[col]].index:
                ret_type[i] = ty
        for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Float') & df_type[col]].index:
            ret_type[i] = ty
    for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Int')].index:
        ret_type[i] = pl.Int64
    for i in df_type.loc[~df_type.index.isin(ret_type) & (df_type['n_unique'] <= cat_max) & (~df_type.index.isin(txt_cols))].index:
        ret_type[i] = pl.Categorical
    for i in df_type.loc[~df_type.index.isin(ret_type)].index:
        ret_type[i] = pl.String
    return ret_type

def get_type_pd(df_type, predefine={}, f32=True, i64=False, cat_max=np.inf, txt_cols = []):
    """
    get data type for pandas DataFrame
    Parameters:
        df_type: pd.DataFrame
            the DataFrame whichi contains type information
        predefine: dict
            the predefined data type
        f32: Boolean
            Use f32 if possible
        i64: Boolean
            use i64 for all integer types
        cat_max: Int
            maximum category number for categorical types
        txt_cols: list
            Text type columns
    Examples:
        >>> df_type = pl.scan_csv('test.csv').pipe(get_type_df)
        >>> dt = get_type_pd(df_type, {'id': 'int64'})
    """
    ret_type = predefine.copy()
    
    for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Date')].index:
        ret_type = 'datetime'
    if f32:
        for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Int') & (df_type['na'] > 0) & (df_type['f32'])].index:
            ret_type[i] = 'float32'
        for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Float') & (df_type['f32'])].index:
            ret_type[i] = 'float32'
    for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Int') & (df_type['na'] > 0)].index:
        ret_type[i] = 'float32'
    for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Float')].index:
        ret_type[i] = 'float32'
    if not i64:
        for col, ty in [('i8', 'int8'), ('i16', 'int16'), ('i32', 'int32')]:
            for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Int') & df_type[col]].index:
                ret_type[i] = ty
        for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Float') & df_type[col]].index:
            ret_type[i] = ty
    for i in df_type.loc[~df_type.index.isin(ret_type) & df_type['dtype'].str.startswith('Int')].index:
        ret_type[i] = 'int64'
    for i in df_type.loc[~df_type.index.isin(ret_type) & (df_type['n_unique'] <= cat_max) & (~df_type.index.isin(txt_cols))].index:
        ret_type[i] = 'category'
    for i in df_type.loc[~df_type.index.isin(ret_type)].index:
        ret_type[i] = 'string'
    return ret_type

def join_and_assign(df1, df2):
    """
    Joins columns from the first DataFrame to the second DataFrame if they do not already exist in the second DataFrame.

    Args:
        df1 (pd.DataFrame): The source DataFrame containing columns to merge.
        df2 (pd.DataFrame): The target DataFrame to which columns from `df1` will be joined if not present.

    Returns:
        pd.DataFrame: The resulting DataFrame with `df1`'s columns added to `df2` where they were missing.
    """
    to_merge = [i for i in df1.columns if i not in df2.columns]
    if len(to_merge) == 0: return df2.copy()
    return df1[to_merge].join(df2)

def combine_cat(df, delimiter=''):
    """
    Combines multiple categorical columns in a DataFrame into a single categorical variable, in efficient way. 

    Parameters:
        df (pd.DataFrame): DataFrame where each column is of categorical dtype.
        delimiter (str): Delimiter

    Returns:
        pd.Series: A Series containing a new categorical variable that represents 
                   the unique combination of all input categorical columns.
    """
    return pd.Series(
        pd.Categorical.from_codes(
            df.apply(lambda x: x.cat.codes).dot(df.nunique().shift(1).fillna(1).astype('int').cumprod()), 
            [delimiter.join(i[::-1]) for i in product(*df[df.columns[::-1]].apply(lambda x: x.cat.categories.astype('str').tolist(), result_type='reduce'))]
        ), index=df.index
    )

def replace_cat(s, rule):
    """
    Replaces the categories in a pandas Categorical Series based on a given rule.

    Args:
        s (pd.Series): A pandas Series with categorical dtype (i.e., `pd.Categorical`).
        rule (Union[Dict[str, str], Callable[[str], str]]): A mapping or function that defines the 
            new category replacements.
            - If `rule` is a dictionary, the keys are the original categories and the values are the 
              replacements.
            - If `rule` is a function, it takes an original category and returns the new category.

    Returns:
        pd.Series: A new pandas Categorical object with the categories replaced 
        according to the given rule.

    Example:
        >>> s = pd.Series(['a', 'b', 'c'], dtype='category')
        >>> rule = {'a': 'x', 'b': 'x'}
        >>> replace_cat(s, rule)
        [x, x, c]
        Categories (2, object): [x, c]

        Or using a function:

        >>> rule = lambda x: x.upper()
        >>> replace_cat(s, rule)
        [A, B, C]
        Categories (3, object): [A, B, C]

    """
    code_replace = {}
    d = {}
    for c, n in zip(
        range(len(s.cat.categories)), 
        s.cat.categories
    ):
        new_cat = rule.get(n, n) if type(rule) == dict else rule(n)
        if new_cat in code_replace:
            d[c] = code_replace[new_cat]
        else:
            d[c] = len(code_replace)
            code_replace[new_cat] = len(code_replace)
    if s.isna().sum() > 0:
        return pd.Series(s.loc[s.notna()].pipe(
                lambda x: pd.Series(
                    pd.Categorical.from_codes(
                        x.cat.codes.map(d), list(code_replace.keys()), ordered=x.cat.ordered
                    ), index = x.index
                )
            ), index=s.index)
    else:
        return pd.Series(
            pd.Categorical.from_codes(
                s.cat.codes.map(d), list(code_replace.keys()), ordered=s.cat.ordered
            ), index=s.index
        )

def rearrange_cat(s_cat, cat_type, repl_rule, use_set=False):
    """
    Rearranges the categories of a pandas Categorical series based on a provided category type
    and a replacement rule for missing categories.

    Args:
        s_cat (pd.Series): A pandas Series of categorical values that need to be rearranged.
        cat_type (pd.api.types.CategoricalDtype): The target CategoricalDtype defining the desired
            order and structure of categories.
        repl_rule (callable): A function that defines a rule for handling categories in `s_cat`
            that are not found in `cat_type`. The function should take two arguments:
            `cat_vals` (the array of category values from `cat_type`) and `x` (a missing category
            from `s_cat`). It returns the value to replace the missing category with.
        use_set(Boolean): provide cat_vals with set
    Returns:
        pd.Categorical: A new pandas Categorical object with the categories of `s_cat` rearranged
        according to `cat_type`, and with missing categories replaced based on `repl_rule`.

    Example:
        >>> cat_type = pd.api.types.CategoricalDtype(categories=["a", "b", "c"])
        >>> s_cat = pd.Series(pd.Categorical(["a", "d", "b"]))
        >>> def repl_rule(cat_vals, x):
        ...     return 0  # Default to the first category if not found
        >>> rearrange_cat(s_cat, cat_type, repl_rule)
        [a, a, b]
        Categories (3, object): [a, b, c]
    """
    cat_vals = cat_type.categories.values
    s_map = pd.Series(np.arange(len(cat_vals)), cat_vals)
    if use_set:
        cat_vals_s = set(cat_vals)
        s_cat_map = pd.Series(s_cat.cat.categories.values, s_cat.cat.categories.values).apply(
            lambda x: s_map[x] if x in s_map else repl_rule(cat_vals_s, x)
        )
    else:
        s_cat_map = pd.Series(s_cat.cat.categories.values, s_cat.cat.categories.values).apply(
            lambda x: s_map[x] if x in s_map else repl_rule(cat_vals, x)
        )
    notna = s_cat.notna()
    return s_cat.loc[notna].pipe(
        lambda x: pd.Series(pd.Series(pd.Categorical.from_codes(x.map(s_cat_map), cat_vals), index=x.index), index=s_cat.index)
    ) if notna.sum() != len(s_cat) else pd.Series(pd.Categorical.from_codes(s_cat.map(s_cat_map), cat_vals), index=s_cat.index)

def split_preprocessor_var(s_names, org_names):
    return s_names.str.split('__|_').apply(
        lambda x: (x, [i for i in range(len(x), 0, -1) if '_'.join(x[1:i]) in org_names][0])
    ).apply(
        lambda x: pd.Series(['_'.join(x[0][:x[1]]), ''.join(x[0][x[1]:])], index=['var1', 'var2'])
    )



"""
Model adapters for handling eval_set in different ML frameworks
"""

from ._base import ModelAdapter
from ._catboost import CatBoostAdapter
from ._keras import KerasAdapter
from ._default import DefaultAdapter
from ._sklearn import LMAdapter, PCAAdapter, LDAAdapter, DecisionTreeAdapter

try:
    from ._xgboost import XGBoostAdapter
except ImportError:
    XGBoostAdapter = None

try:
    from ._lightgbm import LightGBMAdapter
except ImportError:
    LightGBMAdapter = None

# Model adapter registry (인스턴스 저장)
# 모델 클래스명을 키로, 해당 어댑터 인스턴스를 값으로 매핑
# 기본 설정: eval_mode='both', verbose=0.1
MODEL_ADAPTERS = {
    'CatBoostClassifier': CatBoostAdapter(),
    'CatBoostRegressor': CatBoostAdapter(),
    'CatBoostRanker': CatBoostAdapter(),

    'KerasClassifier': KerasAdapter(),
    'KerasRegressor': KerasAdapter(),

    'LinearRegression': LMAdapter(),
    'LogisticRegression': LMAdapter(),

    'PCA': PCAAdapter(),

    'LinearDiscriminantAnalysis': LDAAdapter(),

    'DecisionTreeClassifier': DecisionTreeAdapter(),
    'DecisionTreeRegressor': DecisionTreeAdapter(),
}

if XGBoostAdapter is not None:
    MODEL_ADAPTERS.update({
        'XGBClassifier': XGBoostAdapter(),
        'XGBRegressor': XGBoostAdapter(),
        'XGBRFClassifier': XGBoostAdapter(),
        'XGBRFRegressor': XGBoostAdapter(),
    })

if LightGBMAdapter is not None:
    MODEL_ADAPTERS.update({
        'LGBMClassifier': LightGBMAdapter(),
        'LGBMRegressor': LightGBMAdapter(),
        'LGBMRanker': LightGBMAdapter(),
    })


def get_adapter(model_or_name):
    """모델 또는 모델명에 해당하는 어댑터 인스턴스를 반환

    Args:
        model_or_name: Model instance or model class name (str)

    Returns:
        ModelAdapter: Corresponding adapter instance, or DefaultAdapter instance if not found

    Example:
        >>> from xgboost import XGBClassifier
        >>> adapter = get_adapter(XGBClassifier)
        >>> # or
        >>> adapter = get_adapter('XGBClassifier')
    """
    if isinstance(model_or_name, str):
        model_name = model_or_name
    else:
        # model instance or class
        if hasattr(model_or_name, '__name__'):
            model_name = model_or_name.__name__
        else:
            model_name = model_or_name.__class__.__name__

    return MODEL_ADAPTERS.get(model_name, DefaultAdapter())


def register_adapter(model_name, adapter):
    """새로운 어댑터를 레지스트리에 등록

    Args:
        model_name (str): Model class name
        adapter (ModelAdapter): Adapter instance

    Example:
        >>> class MyCustomAdapter(ModelAdapter):
        ...     def get_fit_params(self, X_train, y_train, X_eval=None, y_eval=None, params=None):
        ...         # custom implementation
        ...         return {...}
        >>>
        >>> register_adapter('MyCustomModel', MyCustomAdapter())
    """
    if not isinstance(adapter, ModelAdapter):
        raise TypeError(f"adapter must be an instance of ModelAdapter, got {type(adapter)}")

    MODEL_ADAPTERS[model_name] = adapter


__all__ = [
    'ModelAdapter',
    'XGBoostAdapter',
    'LightGBMAdapter',
    'CatBoostAdapter',
    'KerasAdapter',
    'DefaultAdapter',
    'LMAdapter',
    'PCAAdapter',
    'LDAAdapter',
    'DecisionTreeAdapter',
    'MODEL_ADAPTERS',
    'get_adapter',
    'register_adapter',
]



import pytest
import pandas as pd
from mllabs._pipeline import Pipeline, PipelineGroup, PipelineNode


class DummyStage:
    __name__ = 'DummyStage'

class DummyHead:
    __name__ = 'DummyHead'

class AnotherProcessor:
    __name__ = 'AnotherProcessor'


@pytest.fixture
def p():
    return Pipeline()


@pytest.fixture
def sp():
    p = Pipeline()
    p.set_grp('stage1', role='stage', processor=DummyStage, method='transform',
              edges={'X': [(None, None)]})
    p.set_node('s1', grp='stage1')
    p.set_grp('head1', role='head', processor=DummyHead, method='predict',
              edges={'X': [(None, None)], 'y': [(None, 'target')]})
    p.set_node('h1', grp='head1')
    return p


class TestInit:
    def test_datasource_exists(self, p):
        assert None in p.nodes
        assert p.nodes[None].name == 'Data_Source'

    def test_empty_grps(self, p):
        assert p.grps == {}


class TestSetGrp:
    def test_new_stage(self, p):
        r = p.set_grp('g1', role='stage')
        assert r['result'] == 'new'
        assert 'g1' in p.grps
        assert p.grps['g1'].role == 'stage'

    def test_new_head(self, p):
        p.set_grp('g1', role='head')
        assert p.grps['g1'].role == 'head'

    def test_with_parent(self, p):
        p.set_grp('parent', role='stage')
        p.set_grp('child', parent='parent')
        assert 'child' in p.grps['parent'].children
        assert p.grps['child'].parent == 'parent'

    def test_parent_role_inherited(self, p):
        p.set_grp('parent', role='head')
        p.set_grp('child', parent='parent')
        assert p.grps['child'].role == 'head'

    def test_parent_not_found(self, p):
        with pytest.raises(ValueError):
            p.set_grp('g1', role='stage', parent='no_exist')

    def test_invalid_role(self, p):
        with pytest.raises(ValueError):
            p.set_grp('g1', role='invalid')

    def test_role_required(self, p):
        with pytest.raises(ValueError):
            p.set_grp('g1')

    def test_name_conflicts_with_node(self, sp):
        with pytest.raises(ValueError):
            sp.set_grp('s1', role='stage')

    def test_exist_skip(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage)
        r = p.set_grp('g1', role='stage', processor=AnotherProcessor, exist='skip')
        assert r['result'] == 'skip'
        assert p.grps['g1'].processor == DummyStage

    def test_exist_error(self, p):
        p.set_grp('g1', role='stage')
        with pytest.raises(ValueError):
            p.set_grp('g1', role='stage', exist='error')

    def test_exist_replace(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage)
        r = p.set_grp('g1', role='stage', processor=AnotherProcessor, exist='replace')
        assert r['result'] == 'update'
        assert p.grps['g1'].processor == AnotherProcessor

    def test_replace_role_change_fails(self, p):
        p.set_grp('g1', role='stage')
        with pytest.raises(ValueError):
            p.set_grp('g1', role='head', exist='replace')

    def test_replace_parent_change(self, p):
        p.set_grp('p1', role='stage')
        p.set_grp('p2', role='stage')
        p.set_grp('child', role='stage', parent='p1')
        assert 'child' in p.grps['p1'].children
        p.set_grp('child', role='stage', parent='p2', exist='replace')
        assert 'child' not in p.grps['p1'].children
        assert 'child' in p.grps['p2'].children

    def test_replace_params_merge(self, p):
        p.set_grp('g1', role='stage', params={'a': 1})
        p.set_grp('g1', role='stage', params={'b': 2}, exist='replace')
        assert p.grps['g1'].params == {'a': 1, 'b': 2}

    def test_with_all_attrs(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage,
                  edges={'X': [(None, None)]}, method='transform',
                  params={'n': 10})
        g = p.grps['g1']
        assert g.processor == DummyStage
        assert g.edges == {'X': [(None, None)]}
        assert g.method == 'transform'
        assert g.params == {'n': 10}


class TestSetNode:
    def test_new_node(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        r = p.set_node('n1', grp='g1')
        assert r['result'] == 'new'
        assert 'n1' in p.nodes
        assert 'n1' in p.grps['g1'].nodes

    def test_grp_not_found(self, p):
        with pytest.raises(ValueError):
            p.set_node('n1', grp='no_exist')

    def test_name_conflicts_with_group(self, p):
        p.set_grp('g1', role='stage')
        with pytest.raises(ValueError):
            p.set_node('g1', grp='g1')

    def test_processor_required(self, p):
        p.set_grp('g1', role='stage', method='transform', edges={'X': [(None, None)]})
        with pytest.raises(ValueError, match='processor'):
            p.set_node('n1', grp='g1')

    def test_method_required(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, edges={'X': [(None, None)]})
        with pytest.raises(ValueError, match='method'):
            p.set_node('n1', grp='g1')

    def test_edges_required(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform')
        with pytest.raises(ValueError, match='edges'):
            p.set_node('n1', grp='g1')

    def test_output_edges_updated(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [('s1', None)]})
        sp.set_node('s2', grp='stage2')
        assert 's2' in sp.nodes['s1'].output_edges

    def test_exist_skip(self, sp):
        r = sp.set_node('s1', grp='stage1', processor=AnotherProcessor, exist='skip')
        assert r['result'] == 'skip'

    def test_exist_error(self, sp):
        with pytest.raises(ValueError):
            sp.set_node('s1', grp='stage1', exist='error')

    def test_exist_replace(self, sp):
        sp.set_node('s1', grp='stage1', processor=AnotherProcessor, exist='replace')
        assert sp.nodes['s1'].processor == AnotherProcessor

    def test_replace_changes_group(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('n1', grp='g1')
        p.set_node('n1', grp='g2', exist='replace')
        assert 'n1' not in p.grps['g1'].nodes
        assert 'n1' in p.grps['g2'].nodes
        assert p.nodes['n1'].grp == 'g2'

    def test_replace_returns_affected_nodes(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [('s1', None)]})
        sp.set_node('s2', grp='stage2')
        r = sp.set_node('s1', grp='stage1', exist='replace')
        assert 's2' in r['affected_nodes']

    def test_replace_preserves_output_edges(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [('s1', None)]})
        sp.set_node('s2', grp='stage2')
        sp.set_node('s1', grp='stage1', exist='replace')
        assert 's2' in sp.nodes['s1'].output_edges

    def test_with_node_level_params(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]}, params={'a': 1})
        p.set_node('n1', grp='g1', params={'b': 2})
        assert p.nodes['n1'].params == {'b': 2}


class TestGroupHierarchy:
    def test_edges_extend(self, p):
        p.set_grp('parent', role='stage', edges={'X': [(None, 'a')]})
        p.set_grp('child', role='stage', parent='parent', edges={'X': [(None, 'b')]})
        attrs = p.grps['child'].get_attrs(p.grps)
        assert attrs['edges']['X'] == [(None, 'b'), (None, 'a')]

    def test_params_no_override(self, p):
        p.set_grp('parent', role='stage', params={'a': 1, 'b': 2})
        p.set_grp('child', role='stage', parent='parent', params={'b': 3, 'c': 4})
        attrs = p.grps['child'].get_attrs(p.grps)
        assert attrs['params'] == {'a': 1, 'b': 3, 'c': 4}

    def test_processor_inherited(self, p):
        p.set_grp('parent', role='stage', processor=DummyStage)
        p.set_grp('child', role='stage', parent='parent')
        attrs = p.grps['child'].get_attrs(p.grps)
        assert attrs['processor'] == DummyStage

    def test_processor_overridden(self, p):
        p.set_grp('parent', role='stage', processor=DummyStage)
        p.set_grp('child', role='stage', parent='parent', processor=AnotherProcessor)
        attrs = p.grps['child'].get_attrs(p.grps)
        assert attrs['processor'] == AnotherProcessor

    def test_method_inherited(self, p):
        p.set_grp('parent', role='stage', method='transform')
        p.set_grp('child', role='stage', parent='parent')
        attrs = p.grps['child'].get_attrs(p.grps)
        assert attrs['method'] == 'transform'

    def test_three_level_hierarchy(self, p):
        p.set_grp('gp', role='stage', processor=DummyStage, edges={'X': [(None, 'a')]},
                  params={'x': 1})
        p.set_grp('par', role='stage', parent='gp', method='transform',
                  edges={'X': [(None, 'b')]}, params={'y': 2})
        p.set_grp('child', role='stage', parent='par',
                  edges={'X': [(None, 'c')]}, params={'z': 3})
        attrs = p.grps['child'].get_attrs(p.grps)
        assert attrs['processor'] == DummyStage
        assert attrs['method'] == 'transform'
        assert attrs['edges']['X'] == [(None, 'c'), (None, 'b'), (None, 'a')]
        assert attrs['params'] == {'x': 1, 'y': 2, 'z': 3}

    def test_attrs_caching(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage)
        attrs1 = p.grps['g1'].get_attrs(p.grps)
        attrs2 = p.grps['g1'].get_attrs(p.grps)
        assert attrs1 is attrs2

    def test_update_attrs_invalidates_cache(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage)
        p.grps['g1'].get_attrs(p.grps)
        assert p.grps['g1'].attrs is not None
        p.grps['g1'].update_attrs()
        assert p.grps['g1'].attrs is None


class TestNodeAttrs:
    def test_merges_from_group(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]}, params={'a': 1})
        p.set_node('n1', grp='g1')
        attrs = p.get_node_attrs('n1')
        assert attrs['processor'] == DummyStage
        assert attrs['method'] == 'transform'
        assert attrs['params'] == {'a': 1}

    def test_node_overrides_processor(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('n1', grp='g1', processor=AnotherProcessor)
        attrs = p.get_node_attrs('n1')
        assert attrs['processor'] == AnotherProcessor

    def test_node_edges_extend_group(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, 'a')]})
        p.set_node('n1', grp='g1', edges={'X': [(None, 'b')]})
        attrs = p.get_node_attrs('n1')
        assert attrs['edges']['X'] == [(None, 'b'), (None, 'a')]

    def test_node_params_no_override(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]}, params={'a': 1, 'b': 2})
        p.set_node('n1', grp='g1', params={'b': 3, 'c': 4})
        attrs = p.get_node_attrs('n1')
        assert attrs['params'] == {'a': 1, 'b': 3, 'c': 4}

    def test_adapter_auto_detect(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('n1', grp='g1')
        attrs = p.get_node_attrs('n1')
        assert attrs['adapter'] is not None

    def test_node_attrs_caching(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('n1', grp='g1')
        a1 = p.nodes['n1'].get_attrs(p.grps)
        a2 = p.nodes['n1'].get_attrs(p.grps)
        assert a1 is a2


class TestNameValidation:
    @pytest.mark.parametrize('name', [
        'test__name', 'a/b', 'a\\b', 'a\0b', 'a<b', 'a>b',
        'a:b', 'a"b', 'a|b', 'a?b', 'a*b'
    ])
    def test_invalid_names_rejected(self, p, name):
        with pytest.raises(ValueError):
            p.set_grp(name, role='stage')

    def test_valid_names_accepted(self, p):
        for name in ['test', 'test_name', 'test-name', 'test123']:
            p.set_grp(name, role='stage')
            assert name in p.grps


class TestEdgeValidation:
    def test_edge_node_not_found(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform')
        with pytest.raises(ValueError, match='not found'):
            p.set_node('n1', grp='g1', edges={'X': [('no_exist', None)]})

    def test_edge_must_be_stage(self, sp):
        with pytest.raises(ValueError, match='stage'):
            sp.set_grp('g2', role='head', edges={'X': [('h1', None)]})


class TestCycleDetection:
    def test_direct_cycle(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [('s1', None)]})
        sp.set_node('s2', grp='stage2')
        with pytest.raises(ValueError, match='cycle'):
            sp.set_node('s1', grp='stage1', edges={'X': [('s2', None)]}, exist='replace')

    def test_indirect_cycle(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('b', grp='g2')
        p.set_grp('g3', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('b', None)]})
        p.set_node('c', grp='g3')
        with pytest.raises(ValueError, match='cycle'):
            p.set_node('a', grp='g1', edges={'X': [('c', None)]}, exist='replace')

    def test_no_cycle_chain(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('b', grp='g2')
        p.set_grp('g3', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('b', None)]})
        p.set_node('c', grp='g3')
        assert 'c' in p.nodes

    def test_diamond_no_cycle(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('b', grp='g2')
        p.set_grp('g3', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('c', grp='g3')
        p.set_grp('g4', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('b', None), ('c', None)]})
        p.set_node('d', grp='g4')
        assert 'd' in p.nodes


class TestRenameGrp:
    def test_basic_rename(self, p):
        p.set_grp('old', role='stage')
        p.rename_grp('old', 'new')
        assert 'old' not in p.grps
        assert 'new' in p.grps
        assert p.grps['new'].name == 'new'

    def test_updates_parent_children(self, p):
        p.set_grp('parent', role='stage')
        p.set_grp('old', role='stage', parent='parent')
        p.rename_grp('old', 'new')
        assert 'new' in p.grps['parent'].children
        assert 'old' not in p.grps['parent'].children

    def test_updates_child_parent(self, p):
        p.set_grp('old', role='stage')
        p.set_grp('child', role='stage', parent='old')
        p.rename_grp('old', 'new')
        assert p.grps['child'].parent == 'new'

    def test_updates_node_grp(self, sp):
        sp.rename_grp('stage1', 'renamed')
        assert sp.nodes['s1'].grp == 'renamed'

    def test_invalidates_node_cache(self, sp):
        sp.nodes['s1'].get_attrs(sp.grps)
        sp.rename_grp('stage1', 'renamed')
        assert sp.nodes['s1'].attrs is None

    def test_source_not_found(self, p):
        with pytest.raises(ValueError):
            p.rename_grp('no_exist', 'new')

    def test_target_exists(self, p):
        p.set_grp('a', role='stage')
        p.set_grp('b', role='stage')
        with pytest.raises(ValueError):
            p.rename_grp('a', 'b')


class TestRemoveGrp:
    def test_remove_empty(self, p):
        p.set_grp('g1', role='stage')
        p.remove_grp('g1')
        assert 'g1' not in p.grps

    def test_updates_parent(self, p):
        p.set_grp('parent', role='stage')
        p.set_grp('child', role='stage', parent='parent')
        p.remove_grp('child')
        assert 'child' not in p.grps['parent'].children

    def test_not_found(self, p):
        with pytest.raises(ValueError):
            p.remove_grp('no_exist')

    def test_has_children(self, p):
        p.set_grp('parent', role='stage')
        p.set_grp('child', role='stage', parent='parent')
        with pytest.raises(ValueError, match='child'):
            p.remove_grp('parent')

    def test_has_nodes(self, sp):
        with pytest.raises(ValueError, match='node'):
            sp.remove_grp('stage1')


class TestRemoveNode:
    def test_remove_leaf(self, sp):
        sp.remove_node('h1')
        assert 'h1' not in sp.nodes
        assert 'h1' not in sp.grps['head1'].nodes

    def test_updates_output_edges(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [('s1', None)]})
        sp.set_node('s2', grp='stage2')
        assert 's2' in sp.nodes['s1'].output_edges
        sp.remove_node('s2')
        assert 's2' not in sp.nodes['s1'].output_edges

    def test_not_found(self, p):
        with pytest.raises(ValueError):
            p.remove_node('no_exist')

    def test_cannot_remove_datasource(self, p):
        with pytest.raises(ValueError):
            p.remove_node(None)

    def test_has_descendants(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [('s1', None)]})
        sp.set_node('s2', grp='stage2')
        with pytest.raises(ValueError, match='dependent'):
            sp.remove_node('s1')


class TestGetNodeNames:
    def test_none_returns_all(self, sp):
        names = sp.get_node_names(None)
        assert None in names
        assert 's1' in names
        assert 'h1' in names

    def test_list_filter(self, sp):
        names = sp.get_node_names(['s1', 'no_exist'])
        assert names == ['s1']

    def test_regex(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [(None, None)]})
        sp.set_node('s2', grp='stage2')
        names = sp.get_node_names('s\\d')
        assert 's1' in names
        assert 's2' in names
        assert 'h1' not in names

    def test_regex_excludes_none(self, sp):
        names = sp.get_node_names('.*')
        assert None not in names

    def test_invalid_type(self, sp):
        with pytest.raises(ValueError):
            sp.get_node_names(123)


class TestGetAffectedNodes:
    def test_chain(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('b', grp='g2')
        p.set_grp('g3', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('b', None)]})
        p.set_node('c', grp='g3')
        result = p._get_affected_nodes(['a'])
        assert result.index('a') < result.index('b') < result.index('c')

    def test_diamond(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('b', grp='g2')
        p.set_grp('g3', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('c', grp='g3')
        p.set_grp('g4', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('b', None), ('c', None)]})
        p.set_node('d', grp='g4')
        result = p._get_affected_nodes(['a'])
        assert result.index('a') < result.index('d')
        assert result.index('b') < result.index('d')
        assert result.index('c') < result.index('d')

    def test_excludes_none(self, sp):
        result = sp._get_affected_nodes(['s1'])
        assert None not in result

    def test_leaf_node(self, sp):
        result = sp._get_affected_nodes(['h1'])
        assert result == ['h1']


class TestCopy:
    def test_independent_copy(self, sp):
        cp = sp.copy()
        assert set(cp.nodes.keys()) == set(sp.nodes.keys())
        assert set(cp.grps.keys()) == set(sp.grps.keys())
        cp.set_grp('new_grp', role='stage')
        assert 'new_grp' not in sp.grps

    def test_preserves_output_edges(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [('s1', None)]})
        sp.set_node('s2', grp='stage2')
        cp = sp.copy()
        assert 's2' in cp.nodes['s1'].output_edges


class TestCopyStage:
    def test_excludes_head(self, sp):
        cp = sp.copy_stage()
        assert 's1' in cp.nodes
        assert 'h1' not in cp.nodes
        assert 'stage1' in cp.grps
        assert 'head1' not in cp.grps

    def test_adjusts_output_edges(self, sp):
        sp.set_grp('stage2', role='stage', processor=DummyStage, method='transform',
                   edges={'X': [('s1', None)]})
        sp.set_node('s2', grp='stage2')
        cp = sp.copy_stage()
        assert 'h1' not in cp.nodes['s1'].output_edges if cp.nodes['s1'].output_edges else True

    def test_datasource_preserved(self, sp):
        cp = sp.copy_stage()
        assert None in cp.nodes


class TestCopyNodes:
    def test_includes_dependencies(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('b', grp='g2')
        p.set_grp('g3', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('b', None)]})
        p.set_node('c', grp='g3')
        cp = p.copy_nodes(['c'])
        assert 'a' in cp.nodes
        assert 'b' in cp.nodes
        assert 'c' in cp.nodes

    def test_excludes_unrelated(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('b', grp='g2')
        cp = p.copy_nodes(['a'])
        assert 'a' in cp.nodes
        assert 'b' not in cp.nodes

    def test_includes_required_groups(self, p):
        p.set_grp('parent', role='stage')
        p.set_grp('child', role='stage', parent='parent', processor=DummyStage,
                  method='transform', edges={'X': [(None, None)]})
        p.set_node('n1', grp='child')
        cp = p.copy_nodes(['n1'])
        assert 'child' in cp.grps
        assert 'parent' in cp.grps

    def test_adjusts_output_edges(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        p.set_grp('g2', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('b', grp='g2')
        p.set_grp('g3', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [('a', None)]})
        p.set_node('c', grp='g3')
        cp = p.copy_nodes(['b'])
        assert 'c' not in cp.nodes['a'].output_edges

    def test_empty_list(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('a', grp='g1')
        cp = p.copy_nodes([])
        assert len(cp.nodes) == 1
        assert None in cp.nodes

    def test_datasource_preserved(self, sp):
        cp = sp.copy_nodes(['s1'])
        assert None in cp.nodes


class TestCompareNodes:
    def test_param_differences(self, p):
        p.set_grp('g1', role='head', processor=DummyHead, method='predict',
                  edges={'X': [(None, None)], 'y': [(None, 'target')]})
        p.set_node('n1', grp='g1', params={'a': 1, 'b': 2})
        p.set_node('n2', grp='g1', params={'a': 1, 'b': 3})
        result = p.compare_nodes(['n1', 'n2'])
        df = result['DummyHead']
        assert ('params', 'b') in df.columns
        assert ('params', 'a') not in df.columns

    def test_groups_by_processor(self, p):
        p.set_grp('g1', role='head', processor=DummyHead, method='predict',
                  edges={'X': [(None, None)], 'y': [(None, 'target')]})
        p.set_node('n1', grp='g1', params={'a': 1})
        p.set_grp('g2', role='head', processor=AnotherProcessor, method='predict',
                  edges={'X': [(None, None)], 'y': [(None, 'target')]})
        p.set_node('n2', grp='g2', params={'a': 2})
        result = p.compare_nodes(['n1', 'n2'])
        assert 'DummyHead' in result
        assert 'AnotherProcessor' in result

    def test_edge_differences(self, p):
        p.set_grp('g1', role='stage', processor=DummyStage, method='transform',
                  edges={'X': [(None, None)]})
        p.set_node('s1', grp='g1')
        p.set_grp('g2', role='head', processor=DummyHead, method='predict',
                  edges={'y': [(None, 'target')]})
        p.set_node('n1', grp='g2', edges={'X': [('s1', ['a', 'b'])]})
        p.set_node('n2', grp='g2', edges={'X': [('s1', ['a', 'c'])]})
        result = p.compare_nodes(['n1', 'n2'])
        df = result['DummyHead']
        x_cols = [c for c in df.columns if c[0] == 'X']
        assert len(x_cols) > 0

    def test_identical_nodes_empty_columns(self, p):
        p.set_grp('g1', role='head', processor=DummyHead, method='predict',
                  edges={'X': [(None, None)], 'y': [(None, 'target')]})
        p.set_node('n1', grp='g1', params={'a': 1})
        p.set_node('n2', grp='g1', params={'a': 1})
        result = p.compare_nodes(['n1', 'n2'])
        df = result['DummyHead']
        assert len(df.columns) == 0


class TestGetParents:
    def test_node_parents(self, p):
        p.set_grp('gp', role='stage')
        p.set_grp('par', role='stage', parent='gp')
        p.set_grp('child', role='stage', parent='par', processor=DummyStage,
                  method='transform', edges={'X': [(None, None)]})
        p.set_node('n1', grp='child')
        result = p.get_parents('n1')
        assert result == ['child', 'par', 'gp']

    def test_datasource(self, p):
        assert p.get_parents(None) == []

    def test_not_found(self, p):
        assert p.get_parents('no_exist') == []



from ._base import ModelAdapter

import pandas as pd
import numpy as np

class LMAdapter(ModelAdapter):
    @staticmethod
    def _get_coef( processor):
        coef_ = processor.obj.coef_
        if len(coef_.shape) == 1:
            if hasattr(processor.obj, 'intercept_'):
                coef_ = np.expand_dims(np.concatenate([coef_, [processor.obj.intercept_]]), axis=0)
                coef_name = list(processor.X_) + ['intercept']
            else:
                coef_ = np.expand_dims(processor.obj.coef_, axis=0)
                coef_name = processor.X_
            idx = [0]
        else:
            if hasattr(processor.obj, 'intercept_'):
                coef_ = np.concatenate([coef_, np.expand_dims(processor.obj.intercept_, axis=0)], axis=1)
                coef_name = list(processor.X_) + ['intercept']
            else:
                coef_ = processor.obj.coef_
                coef_name = processor.X_
            idx = np.arange(coef_.shape[0])
        return pd.DataFrame(coef_, index=idx, columns=coef_name).stack().rename('coef')

LMAdapter.result_objs = {'coef': (LMAdapter._get_coef, True)}

class PCAAdapter(ModelAdapter):
    @staticmethod
    def _get_explained_variance(processor):
        obj = processor.obj
        output_vars = list(processor.output_vars) if hasattr(processor, 'output_vars') and processor.output_vars is not None else list(range(len(obj.explained_variance_)))
        return pd.Series(
            obj.explained_variance_, index=output_vars, name = 'variance'
        )
    
    @staticmethod
    def _get_explained_variance_ratio(processor):
        obj = processor.obj
        output_vars = list(processor.output_vars) if hasattr(processor, 'output_vars') and processor.output_vars is not None else list(range(len(obj.explained_variance_ratio_)))
        return pd.Series(
            obj.explained_variance_ratio_, index=output_vars, name='ratio'
        )

    @staticmethod
    def _get_components(processor):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.components_.shape[1]))
        output_vars = list(processor.output_vars) if hasattr(processor, 'output_vars') and processor.output_vars is not None else list(range(obj.components_.shape[0]))
        return pd.DataFrame(
            obj.components_,
            index=output_vars,
            columns=input_vars
        ).stack().rename('component')

PCAAdapter.result_objs = {
    'explained_variance': (PCAAdapter._get_explained_variance, True),
    'explained_variance_ratio': (PCAAdapter._get_explained_variance_ratio, True),
    'components': (PCAAdapter._get_components, True)
}


class LDAAdapter(ModelAdapter):
    @staticmethod
    def _get_coef(processor):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.coef_.shape[-1]))
        classes = list(processor.classes_) if hasattr(processor, 'classes_') and processor.classes_ is not None else list(range(obj.coef_.shape[0] if obj.coef_.ndim > 1 else 1))

        coef = obj.coef_
        if coef.ndim == 1:
            coef = coef.reshape(1, -1)

        return pd.DataFrame(
            coef,
            index=classes,
            columns=input_vars
        ).unstack().rename('value')

    @staticmethod
    def _get_intercept(processor):
        obj = processor.obj
        classes = list(processor.classes_) if hasattr(processor, 'classes_') and processor.classes_ is not None else list(range(len(np.atleast_1d(obj.intercept_))))

        intercept = obj.intercept_
        if np.isscalar(intercept):
            intercept = [intercept]

        return pd.Series(
            intercept, index=classes, name = 'intercept'
        )

    @staticmethod
    def _get_scalings(processor):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.scalings_.shape[0]))
        output_vars = list(processor.output_vars) if hasattr(processor, 'output_vars') and processor.output_vars is not None else [f'LD{i}' for i in range(obj.scalings_.shape[1])]

        return pd.DataFrame(
            obj.scalings_,
            index=input_vars,
            columns=output_vars
        ).unstack().rename('scale')

    @staticmethod
    def _get_explained_variance_ratio(processor):
        obj = processor.obj
        output_vars = list(processor.output_vars) if hasattr(processor, 'output_vars') and processor.output_vars is not None else [f'LD{i}' for i in range(len(obj.explained_variance_ratio_))]

        return pd.Series(
            obj.explained_variance_ratio_, index=output_vars, name = 'ratio'
        )
LDAAdapter.result_objs = {
    'coef': (LDAAdapter._get_coef, True),
    'intercept': (LDAAdapter._get_intercept, True),
    'scalings': (LDAAdapter._get_scalings, True),
    'explained_variance_ratio': (LDAAdapter._get_explained_variance_ratio, True)
}

class DecisionTreeAdapter(ModelAdapter):
    @staticmethod
    def _get_feature_importances(processor):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(len(obj.feature_importances_)))

        return pd.Series(
            obj.feature_importances_, index=input_vars, name = 'importance'
        )
    
    @staticmethod
    def _get_tree(processor):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.n_features_in_))

        tree_ = obj.tree_
        tree_structure = []
        for i in range(tree_.node_count):
            feature_idx = tree_.feature[i]
            node_dict = {
                'node_id': i,
                'feature': input_vars[feature_idx] if feature_idx >= 0 else None,
                'threshold': tree_.threshold[i] if feature_idx >= 0 else None,
                'impurity': tree_.impurity[i],
                'n_samples': tree_.n_node_samples[i],
                'left_child': tree_.children_left[i] if tree_.children_left[i] >= 0 else None,
                'right_child': tree_.children_right[i] if tree_.children_right[i] >= 0 else None,
                'value': tree_.value[i].tolist()
            }
            tree_structure.append(node_dict)

        return pd.DataFrame(tree_structure)

    @staticmethod
    def _plot_tree(processor, **args):
        from sklearn.tree import plot_tree
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.n_features_in_))
        plot_tree(obj, feature_names = input_vars, **args)

DecisionTreeAdapter.result_objs = {
    'feature_importances': (DecisionTreeAdapter._get_feature_importances, True),
    'tree': (DecisionTreeAdapter._get_tree, False), 
    'plot_tree': (DecisionTreeAdapter._plot_tree, False), 
}


import pickle
import shap

from ._base import Collector
from .._data_wrapper import unwrap


class SHAPCollector(Collector):
    def __init__(self, name, connector, explainer_cls=None, data_filter=None):
        super().__init__(name, connector)
        self.explainer_cls = explainer_cls or shap.TreeExplainer
        self.data_filter = data_filter
        self._buffer = {}
        self._mem_data = {}

    def _start(self, node):
        self._buffer[node] = {}

    def _collect(self, node, idx, inner_idx, context):
        (train_X, _), valid_X = context['input']['X']

        model = context['processor'].obj
        explainer = self.explainer_cls(model)

        train_dict = {'X': train_X}
        if self.data_filter is not None:
            train_dict = self.data_filter(train_dict)

        valid_dict = {'X': valid_X}
        if self.data_filter is not None:
            valid_dict = self.data_filter(valid_dict)

        self._buffer[node][(idx, inner_idx)] = {
            'train': explainer.shap_values(unwrap(train_dict['X'])),
            'valid': explainer.shap_values(unwrap(valid_dict['X'])),
            'train_index': train_dict['X'].get_index(),
            'valid_index': valid_dict['X'].get_index(),
            'columns': list(context['processor'].X_),
        }

    def _end_idx(self, node, idx):
        entries = {k: v for k, v in self._buffer[node].items() if k[0] == idx}
        if not entries:
            return

        if self.path is not None:
            self._ensure_path()
            file_path = self.path / f"{node}.pkl"
            if file_path.exists():
                with open(file_path, 'rb') as f:
                    existing = pickle.load(f)
                existing.update(entries)
            else:
                existing = dict(entries)
            with open(file_path, 'wb') as f:
                pickle.dump(existing, f)
        else:
            if node not in self._mem_data:
                self._mem_data[node] = {}
            self._mem_data[node].update(entries)

        for k in entries:
            del self._buffer[node][k]

    def _end(self, node):
        if node in self._buffer:
            del self._buffer[node]

    def has_node(self, node):
        if node in self._mem_data:
            return True
        if self.path is not None:
            return (self.path / f"{node}.pkl").exists()
        return False

    def reset_nodes(self, nodes):
        for node in nodes:
            if self.path is not None:
                file_path = self.path / f"{node}.pkl"
                if file_path.exists():
                    file_path.unlink()
            if node in self._mem_data:
                del self._mem_data[node]

    def save(self):
        if self.path is None:
            return
        self._ensure_path()
        config = {
            'name': self.name,
            'connector': self.connector,
            'explainer_cls': self.explainer_cls,
            'data_filter': self.data_filter,
        }
        with open(self.path / '__config.pkl', 'wb') as f:
            pickle.dump(config, f)

    @classmethod
    def load(cls, path):
        with open(path / '__config.pkl', 'rb') as f:
            config = pickle.load(f)
        obj = cls(
            name=config['name'],
            connector=config['connector'],
            explainer_cls=config.get('explainer_cls'),
            data_filter=config.get('data_filter'),
        )
        obj.path = path
        return obj

    def _load_node_data(self, node):
        if self.path is not None:
            file_path = self.path / f"{node}.pkl"
            if not file_path.exists():
                raise FileNotFoundError(f"SHAP data not found: {file_path}")
            with open(file_path, 'rb') as f:
                return pickle.load(f)
        else:
            if node not in self._mem_data:
                raise KeyError(f"SHAP data not found in memory: {node}")
            return self._mem_data[node]

    def _get_saved_nodes(self):
        if self.path is not None:
            if not self.path.exists():
                return []
            return [f.stem for f in self.path.glob("*.pkl") if not f.stem.startswith('__')]
        else:
            return list(self._mem_data.keys())




from abc import ABC, abstractmethod

class BaseLogger(ABC):
    def __init__(self, level = ['info', 'warning', 'progress'], n_store_warning = 1000):
        self._progress = list()
        self.show_info = 'info' in level
        self.show_warning = 'warning' in level
        self.show_progress = 'progress' in level
        self.warning_list = list()
        self.n_store_warning = n_store_warning

    @abstractmethod
    def info(self, msg):
        pass

    def warning(self, msg):
        pass

    def start_progress(self, title, total=None):
        pass

    def update_progress(self, current):
        pass

    def end_progress(self, current=None):
        pass

    def adhoc_progress(self, current, total, msg=None):
        pass

    def clear_progress(self):
        pass

class DefaultLogger(BaseLogger):
    def info(self, msg):
        if self.show_info:
            print(msg)

    def warning(self, msg):
        if self.show_warning:
            print(msg)
        self.warning_list.append(msg)
        self.warning_list = self.warning_list[-self.n_store_warning:]

    def _render_progress(self):
        parts = []
        for title, current, total in self._progress:
            if total is not None and total > 0:
                pct = int(current * 100 / total)
                parts.append(f"{title} {current}/{total} ({pct}%)")
            else:
                parts.append(f"{title} {current}")
        print(f"\r{' > '.join(parts)}", end='', flush=True)

    def start_progress(self, title, total=None):
        self._progress.append([title, 0, total])
        if self.show_progress:
            self._render_progress()

    def update_progress(self, current):
        self._progress[-1][1] = current
        if self.show_progress:
            self._render_progress()

    def end_progress(self, current=None):
        if current is not None:
            self._progress[-1][1] = current
            if self.show_progress:
                self._render_progress()
        self._progress.pop()
        if len(self._progress) == 0:
            print()

    def adhoc_progress(self, current, total, msg=None):
        if not self.show_progress or len(self._progress) == 0:
            return
        parts = []
        for title, cur, tot in self._progress:
            if tot is not None and tot > 0:
                pct = int(cur * 100 / tot)
                parts.append(f"{title} {cur}/{tot} ({pct}%)")
            else:
                parts.append(f"{title} {cur}")
        pct = int(current * 100 / total) if total > 0 else 0
        adhoc_str = f"{current}/{total} ({pct}%)"
        if msg:
            adhoc_str += f" {msg}"
        parts.append(adhoc_str)
        print(f"\r{' > '.join(parts)}", end='', flush=True)

    def clear_progress(self):
        if len(self._progress) > 0:
            self._progress.clear()
            print()



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

try:
    import pandas as pd
except Exception:
    pd = None

try:
    import polars as pl
except Exception:
    pl = None


def _is_nan(x):
    try:
        return isinstance(x, float) and np.isnan(x)
    except Exception:
        return False


class CategoricalPairCombiner(BaseEstimator, TransformerMixin):
    """
    두 범주형 변수를 결합해서 하나의 범주형 변수를 만드는 처리기.

    - pairs: [(col_a, col_b), ...]
      * pandas/polars: str(컬럼명) 또는 int(컬럼 위치) 가능
      * numpy: int(컬럼 위치)만 권장
    - min_frequency:
      * fit에서 결합값 빈도를 세고, 빈도가 min_frequency 이하(<=)인 결합값은 transform 시 결측으로 처리

    반환:
      * pandas/polars: 원본 DF에 결합 컬럼을 추가해서 반환
      * numpy: 원본 배열에 결합 컬럼(들)을 뒤에 추가해서 반환 (dtype=object)
    """

    def __init__(
        self,
        pairs,
        min_frequency,
        *,
        sep="__",
        drop_original=False,
        treat_empty_string_as_missing=True,
        missing_value=None,
        new_col_names=None,
    ):
        self.pairs = list(pairs)
        self.min_frequency = int(min_frequency)
        self.sep = sep
        self.drop_original = drop_original
        self.treat_empty_string_as_missing = treat_empty_string_as_missing
        self.missing_value = missing_value
        self.new_col_names = list(new_col_names) if new_col_names is not None else None

        self._allowed_ = {}
        self._resolved_pairs_ = []
        self._new_names_ = []

    # ----------------- helpers -----------------

    def _is_missing(self, v):
        if v is None:
            return True
        if _is_nan(v):
            return True
        if self.treat_empty_string_as_missing and isinstance(v, str) and v == "":
            return True
        return False

    def _combine(self, a, b):
        if self._is_missing(a) or self._is_missing(b):
            return None
        return f"{a}{self.sep}{b}"

    def _default_new_name(self, a, b):
        return f"{a}{self.sep}{b}"

    def _detect_kind(self, X):
        if pd is not None:
            if isinstance(X, pd.DataFrame):
                return "pandas_df"
        if pl is not None:
            if isinstance(X, pl.DataFrame):
                return "polars_df"
        if isinstance(X, np.ndarray):
            return "numpy"
        raise TypeError(f"Unsupported input type: {type(X)}")

    def _resolve_pair(self, X, p):
        a, b = p

        # pandas/polars: allow int positional
        if isinstance(X, np.ndarray):
            if not (isinstance(a, int) and isinstance(b, int)):
                raise ValueError("For numpy input, pair elements should be integer column indices.")
            return a, b, str(a), str(b)

        # DataFrame case
        cols = list(X.columns)
        if isinstance(a, int):
            a_name = cols[a]
        else:
            a_name = a
        if isinstance(b, int):
            b_name = cols[b]
        else:
            b_name = b
        return a_name, b_name, str(a_name), str(b_name)

    def _make_new_names(self):
        if self.new_col_names is not None:
            if len(self.new_col_names) != len(self._resolved_pairs_):
                raise ValueError("new_col_names length must match pairs length.")
            return list(self.new_col_names)
        # default names: "colA__colB" (sep 사용)
        names = []
        for _, _, a_label, b_label in self._resolved_pairs_:
            names.append(self._default_new_name(a_label, b_label))
        return names

    # ----------------- sklearn API -----------------

    def fit(self, X, y=None):
        kind = self._detect_kind(X)
        self._allowed_.clear()
        self._resolved_pairs_.clear()

        # resolve pairs to concrete column selectors/names
        if kind == "numpy":
            for p in self.pairs:
                a_idx, b_idx, a_label, b_label = self._resolve_pair(X, p)
                self._resolved_pairs_.append((a_idx, b_idx, a_label, b_label))
        else:
            for p in self.pairs:
                a_name, b_name, a_label, b_label = self._resolve_pair(X, p)
                self._resolved_pairs_.append((a_name, b_name, a_label, b_label))

        self._new_names_ = self._make_new_names()

        # count frequencies
        for (a_sel, b_sel, a_label, b_label), new_name in zip(self._resolved_pairs_, self._new_names_):
            freq = {}
            if kind == "numpy":
                arr = np.asarray(X)
                if arr.ndim == 1:
                    arr = arr.reshape(-1, 1)
                col_a = arr[:, a_sel].tolist()
                col_b = arr[:, b_sel].tolist()
                for va, vb in zip(col_a, col_b):
                    comb = self._combine(va, vb)
                    if comb is None:
                        continue
                    freq[comb] = freq.get(comb, 0) + 1
            elif kind == "pandas_df":
                col_a = X[a_sel].tolist()
                col_b = X[b_sel].tolist()
                for va, vb in zip(col_a, col_b):
                    comb = self._combine(va, vb)
                    if comb is None:
                        continue
                    freq[comb] = freq.get(comb, 0) + 1
            elif kind == "polars_df":
                col_a = X.get_column(a_sel).to_list()
                col_b = X.get_column(b_sel).to_list()
                for va, vb in zip(col_a, col_b):
                    comb = self._combine(va, vb)
                    if comb is None:
                        continue
                    freq[comb] = freq.get(comb, 0) + 1
            else:
                raise RuntimeError("Unexpected kind")

            # allowed: 빈도 > min_frequency (<=는 결측 처리)
            allowed = {k for k, c in freq.items() if c > self.min_frequency}
            self._allowed_[new_name] = allowed

        return self

    def transform(self, X):
        if not hasattr(self, "_allowed_") or not self._allowed_:
            raise RuntimeError("This transformer is not fitted yet. Call fit() first.")

        kind = self._detect_kind(X)

        if kind == "pandas_df":
            return self._transform_pandas(X)
        if kind == "polars_df":
            return self._transform_polars(X)
        if kind == "numpy":
            return self._transform_numpy(X)

        raise TypeError(f"Unsupported input type: {type(X)}")

    # ----------------- transforms -----------------

    def _transform_pandas(self, X):
        if pd is None:
            raise RuntimeError("pandas is not installed.")
        X_out = X.copy()

        for (a_name, b_name, _, _), new_name in zip(self._resolved_pairs_, self._new_names_):
            allowed = self._allowed_.get(new_name, set())

            def _map_row(va, vb):
                comb = self._combine(va, vb)
                if comb is None:
                    return self.missing_value
                if comb in allowed:
                    return comb
                return self.missing_value

            X_out[new_name] = [
                _map_row(va, vb) for va, vb in zip(X_out[a_name].tolist(), X_out[b_name].tolist())
            ]

            if self.drop_original:
                X_out = X_out.drop(columns=[a_name, b_name], errors="ignore")

        return X_out

    def _transform_polars(self, X):
        if pl is None:
            raise RuntimeError("polars is not installed.")
        df = X

        exprs = []
        cols_to_drop = []

        for (a_name, b_name, _, _), new_name in zip(self._resolved_pairs_, self._new_names_):
            allowed = self._allowed_.get(new_name, set())

            def _map_struct(s):
                va = s[a_name]
                vb = s[b_name]
                comb = self._combine(va, vb)
                if comb is None:
                    return self.missing_value
                if comb in allowed:
                    return comb
                return self.missing_value

            exprs.append(
                pl.struct([pl.col(a_name), pl.col(b_name)]).map_elements(_map_struct, return_dtype=pl.Utf8).alias(new_name)
            )

            if self.drop_original:
                cols_to_drop.extend([a_name, b_name])

        if exprs:
            df = df.with_columns(exprs)
        if self.drop_original and cols_to_drop:
            # 중복 제거
            cols_to_drop = list(dict.fromkeys(cols_to_drop))
            df = df.drop(cols_to_drop)

        return df

    def _transform_numpy(self, X):
        arr = np.asarray(X)
        orig_ndim = arr.ndim
        if arr.ndim == 1:
            arr = arr.reshape(-1, 1)

        out = arr.astype(object, copy=True)
        new_cols = []

        for (a_idx, b_idx, _, _), new_name in zip(self._resolved_pairs_, self._new_names_):
            allowed = self._allowed_.get(new_name, set())
            col = np.empty((out.shape[0],), dtype=object)

            for i in range(out.shape[0]):
                comb = self._combine(out[i, a_idx], out[i, b_idx])
                if comb is None:
                    col[i] = self.missing_value
                elif comb in allowed:
                    col[i] = comb
                else:
                    col[i] = self.missing_value

            new_cols.append(col.reshape(-1, 1))

        if new_cols:
            out = np.concatenate([out] + new_cols, axis=1)

        if self.drop_original:
            # 원본 컬럼 제거(쌍에서 나온 컬럼들)
            drop_idxs = sorted({a for (a, _, _, _) in self._resolved_pairs_} | {b for (_, b, _, _) in self._resolved_pairs_})
            keep = [j for j in range(out.shape[1] - len(new_cols)) if j not in drop_idxs]
            # keep 원본 + 새 컬럼들
            out = np.concatenate([out[:, keep]] + new_cols, axis=1)

        if orig_ndim == 1:
            # 1D 입력은 기본적으로 2D로 늘리지만, 원하면 그대로 반환하도록 여기서는 그대로 2D 유지가 안전
            return out
        return out

class CatOOVFilter(TransformerMixin, BaseEstimator):
    def __init__(self):
        pass
    def fit(self, X, y = None):
        self.s_dtype_ = {i: X[i].dtype for i in X.columns}
        self.s_mode_ = X.apply(lambda x: x.mode()[0])
        self.fitted_ = True
        return self
    
    def transform(self, X):
        return pd.concat([
            dproc.rearrange_cat(X[k], v, lambda d, c: 0 if c not in d else c, use_set = True).rename(k)
            for k, v in self.s_dtype_.items()
        ], axis=1)
    def get_params(self, deep=True):
        return {
        }

    def set_output(self, transform='pandas'):
        pass

    def get_feature_names_out(self, X = None):
        return list(self.s_dtype_.keys())


import re
from ._data_wrapper import unwrap

def resolve_columns(data, X, y=None, processor=None):
    """X와 y를 실제 컬럼 리스트로 변환"""
    columns = data.get_columns()

    # y 처리 (y가 있으면 X에서 제외할 컬럼)
    y_cols = []
    if y is not None:
        if isinstance(y, slice):
            y_cols = columns[y]
        elif isinstance(y, list):
            y_cols = y
        else:
            y_cols = [y]

    # X 처리
    if X is None:
        # y가 있으면 y를 제외한 모든 컬럼
        if y is not None:
            return [col for col in columns if col not in y_cols]
        else:
            return columns
    elif isinstance(X, str):
        # 정규 표현식 패턴이면 매칭되는 컬럼만 선택
        return [col for col in columns if re.match(X, col)]
    elif callable(X):
        mask = X(columns, processor=processor)
        return [col for col, keep in zip(columns, mask) if keep]
    elif isinstance(X, slice):
        # slice 객체면 컬럼을 슬라이싱
        return columns[X]
    elif isinstance(X, tuple):
        if len(X) == 0:
            return []
        head = X[0]
        if callable(head):
            mask = head(columns, *X[1:], processor=processor)
            return [col for col, keep in zip(columns, mask) if keep]
        elif head == 'cup':
            seen = set()
            ret = list()
            for x in X[1:]:
                for col in resolve_columns(data, x, processor=processor):
                    if col not in seen:
                        seen.add(col)
                        ret.append(col)
            return ret
        elif head == 'cap':
            sets = [set(resolve_columns(data, x, processor=processor)) for x in X[1:]]
            if not sets:
                return []
            common = sets[0]
            for s in sets[1:]:
                common &= s
            first = resolve_columns(data, X[1], processor=processor)
            return [col for col in first if col in common]
        else:
            raise ValueError(f"tuple의 첫번째 요소가 callable이 아닌 경우 'cup' 또는 'cap'이어야 함: {head}")
    elif isinstance(X, list):
        for x in X:
            if x is None or isinstance(x, list):
                raise ValueError(f"list 요소에 list나 None은 허용되지 않음: {x}")
        seen = set()
        ret = list()
        for x in X:
            for col in resolve_columns(data, x, processor=processor):
                if col not in seen:
                    seen.add(col)
                    ret.append(col)
        return ret
    else:
        # 단일 값이면 리스트로 변환
        ret = list()
        for col in columns:
            if re.match(X, col):
                ret.append(col)
                break
        return ret

class TransformProcessor():
    def __init__(self, name, transformer, adapter = None, params = {}, logger = None):
        self.name = name
        self.transformer = transformer
        self.params = params
        self.adapter = adapter
        self.output_vars = None
        self.logger = logger

    def fit(self, data_dict):
        # X key로 데이터 가져오기
        train_X, train_v_X = data_dict['X']
        self.X_ = train_X.get_columns()

        # y key로 데이터 가져오기 (있으면)
        if 'y' in data_dict:
            train_y, train_v_y = data_dict['y']
            self.y_columns = train_y.get_columns()
        else:
            train_y, train_v_y = None, None
            self.y_columns = None

        params = self.adapter.get_params(self.params, logger = self.logger) if self.adapter is not None else self.params
        self.obj = self.transformer(**params)

        # adapter에서 fit_params 생성
        if self.adapter is not None:
            fit_params = self.adapter.get_fit_params(
                data_dict=data_dict, X=self.X_, y=self.y_columns, params=self.params,logger=self.logger
            )
        else:
            fit_params = {}

        # DataWrapper에서 native로 변환
        train_X_native = unwrap(train_X)

        if train_y is None:
            self.obj.fit(train_X_native, **fit_params)
        else:
            train_y_native = unwrap(train_y.squeeze())
            self.obj.fit(train_X_native, train_y_native, **fit_params)

        # 컬럼명 결정 (get_feature_names_out이 있으면 사용)
        if hasattr(self.obj, 'get_feature_names_out'):
            column_names = self.obj.get_feature_names_out().tolist()
            column_names = [f"{self.name}__{col}" for col in column_names]
        else:
            column_names = None

        if column_names is not None:
            self.output_vars = column_names
        return self

    def fit_process(self, data_dict):
        # X key로 데이터 가져오기
        train_X, train_v_X = data_dict['X']
        self.X_ = train_X.get_columns()
        train_index = train_X.get_index()

        # y key로 데이터 가져오기 (있으면)
        if 'y' in data_dict:
            train_y, train_v_y = data_dict['y']
            self.y_columns = train_y.get_columns()
        else:
            train_y, train_v_y = None, None
            self.y_columns = None

        params = self.adapter.get_params(self.params, logger = self.logger) if self.adapter is not None else self.params
        self.obj = self.transformer(**params)

        # adapter에서 fit_params 생성
        if self.adapter is not None:
            fit_params = self.adapter.get_fit_params(
                data_dict=data_dict, X=self.X_, y=self.y_columns, params=self.params, logger=self.logger
            )
        else:
            fit_params = {}

        # DataWrapper에서 native로 변환
        train_X_native = unwrap(train_X)

        if train_y is None:
            result = self.obj.fit_transform(train_X_native, **fit_params)
        else:
            train_y_native = unwrap(train_y.squeeze())
            result = self.obj.fit_transform(train_X_native, train_y_native, **fit_params)

        # train의 Wrapper 타입으로 변환
        train_wrapper_class = type(train_X)
        # 컬럼명 결정 (get_feature_names_out이 있으면 사용)
        if hasattr(self.obj, 'get_feature_names_out'):
            column_names = self.obj.get_feature_names_out().tolist()
            column_names = [f"{self.name}__{col}" for col in column_names]
        else:
            column_names = None

        if column_names is not None:
            self.output_vars = column_names
        return train_wrapper_class.from_output(result, self.output_vars, train_index)

    def process(self, data):
        # DataWrapper에서 native로 변환
        data_X = unwrap(data)
        data_index = data.get_index()

        result = self.obj.transform(data_X)

        # data의 Wrapper 타입으로 변환
        data_wrapper_class = type(data)
        return data_wrapper_class.from_output(result, self.output_vars, data_index)

class PredictProcessor():
    def __init__(self, name, estimator, method='predict', adapter = None, params = {}, logger = None):
        self.name = name
        self.estimator = estimator
        self.params = params
        self.method = method
        self.output_vars = None
        self.adapter = adapter
        self.y_columns = None
        self.logger = logger

    def fit(self, data_dict):
        # X key로 데이터 가져오기
        train_X, train_v_X = data_dict['X']
        self.X_ = train_X.get_columns()

        # y key로 데이터 가져오기 (있으면)
        if 'y' in data_dict:
            train_y, train_v_y = data_dict['y']
            self.y_columns = train_y.get_columns()
        else:
            train_y, train_v_y = None, None
            self.y_columns = None

        # adapter가 있으면 params 조정 (callbacks 등 설정)
        params = self.adapter.get_params(self.params, logger = self.logger) if self.adapter is not None else self.params
        self.obj = self.estimator(**params)

        # adapter에서 fit_params 생성
        if self.adapter is not None:
            fit_params = self.adapter.get_fit_params(
                data_dict=data_dict, X=self.X_, y=self.y_columns, params=self.params, logger=self.logger
            )
        else:
            fit_params = {}

        # DataWrapper에서 native로 변환
        train_X_native = unwrap(train_X)

        if train_y is None:
            # 비지도학습
            self.obj.fit(train_X_native, **fit_params)
        else:
            train_y_native = unwrap(train_y.squeeze())
            # 지도학습
            self.obj.fit(train_X_native, train_y_native, **fit_params)

        if self.method == 'predict':
            # y 변수명 결정
            if self.y_columns is None:
                y_name = 'prediction'
            else:
                y_name = '_'.join(self.y_columns)

            col_name = f"{self.name}__{y_name}"
            self.output_vars = [col_name]
        elif self.method == 'predict_proba':
            # y 변수명 결정
            if self.y_columns is None:
                y_name = 'prediction'
            else:
                y_name = '_'.join(self.y_columns)

            columns = [f"{self.name}__{y_name}_{i}" for i in self.obj.classes_]
            self.output_vars = columns
        return self

    def fit_process(self, data_dict):
        # X key로 데이터 가져오기
        train_X, train_v_X = data_dict['X']
        self.X_ = train_X.get_columns()
        train_index = train_X.get_index()

        # y key로 데이터 가져오기 (있으면)
        if 'y' in data_dict:
            train_y, train_v_y = data_dict['y']
            self.y_columns = train_y.get_columns()
        else:
            train_y, train_v_y = None, None
            self.y_columns = None

        # adapter가 있으면 params 조정 (callbacks 등 설정)
        params = self.adapter.get_params(self.params, logger = self.logger) if self.adapter is not None else self.params
        self.obj = self.estimator(**params)

        # adapter에서 fit_params 생성
        if self.adapter is not None:
            fit_params = self.adapter.get_fit_params(
                data_dict=data_dict, X=self.X_, y=self.y_columns, params=self.params,
                logger=self.logger
            )
        else:
            fit_params = {}

        # DataWrapper에서 native로 변환
        train_X_native = unwrap(train_X)

        if train_y is None:
            # 비지도학습
            predictions = self.obj.fit_predict(train_X_native, **fit_params)
        else:
            # 지도학습
            train_y_native = unwrap(train_y.squeeze())
            predictions = self.obj.fit_predict(train_X_native, train_y_native, **fit_params)

        # 컬럼명 결정
        if self.y_columns is None:
            y_name = 'prediction'
        else:
            y_name = '_'.join(self.y_columns)

        col_name = f"{self.name}__{y_name}"
        column_names = [col_name]
        self.output_vars = column_names

        # train의 Wrapper 타입으로 변환
        train_wrapper_class = type(train_X)
        return train_wrapper_class.from_output(predictions, column_names, train_index)

    def process(self, data):
        # DataWrapper에서 native로 변환
        data_X = unwrap(data)
        data_index = data.get_index()

        if self.method == 'predict':
            predictions = self.obj.predict(data_X)
            # 컬럼명은 fit에서 이미 결정됨
            column_names = self.output_vars

        elif self.method == 'predict_proba':
            if not hasattr(self.obj, 'predict_proba'):
                raise Exception(f"Model {self.estimator.__name__} does not support predict_proba")

            predictions = self.obj.predict_proba(data_X)
            # 컬럼명은 fit에서 이미 결정됨
            column_names = self.output_vars

        else:
            raise ValueError(f"Unknown method: {self.method}. Use 'predict' or 'predict_proba'")

        # data의 Wrapper 타입으로 변환
        data_wrapper_class = type(data)
        return data_wrapper_class.from_output(predictions, column_names, data_index)



"""
XGBoost adapter
"""

import pandas as pd
from ._base import ModelAdapter

from xgboost.callback import TrainingCallback

class ProgressCallback(TrainingCallback):
    def __init__(self, n_estimators, period_pct, logger):
        self.n_estimators = n_estimators
        self.period_pct = period_pct
        self.last_printed = -1
        self.logger = logger

    def after_iteration(self, model, epoch, evals_log):
        current = epoch + 1
        percentage = (current / self.n_estimators) * 100

        # period_pct마다 출력
        if int(percentage / (self.period_pct * 100)) > self.last_printed:
            self.last_printed = int(percentage / (self.period_pct * 100))

            # metric 정보 추출
            metrics_str = ""
            if evals_log:
                last_metrics = []
                for dataset, metrics in evals_log.items():
                    for metric_name, values in metrics.items():
                        last_metrics.append(f"{dataset}-{metric_name}: {values[-1]:.4f}")
                metrics_str = ", ".join(last_metrics)
            self.logger.adhoc_progress(current, self.n_estimators, metrics_str if metrics_str else None)
        return False

class XGBoostAdapter(ModelAdapter):
    """Adapter for XGBoost models (XGBClassifier, XGBRegressor)

    XGBoost는 eval_set 파라미터로 [(X, y), ...] 형태를 받습니다.
    """

    result_objs = [
        'feature_importances_weight', 'feature_importances_gain', 'feature_importances_cover',
        'feature_importances_total_gain', 'feature_importances_total_cover',
        'evals_result', 'trees'
    ]

    def get_params(self, params, logger = None):
        """XGBoost 모델 생성자 파라미터 조정 (ProgressCallback 설정)"""
        if params is None:
            params = {}
        else:
            params = params.copy()
        if self.verbose > 0 and self.verbose < 1:
            # 0 < verbose < 1: 진행률 기반 출력을 위한 callback 설정

            n_estimators = params.get('n_estimators', 100)
            callbacks = params.get('callbacks', [])
            if logger is not None:
                callbacks.append(ProgressCallback(n_estimators, self.verbose, logger))
            params['callbacks'] = callbacks

        return params

    def get_fit_params(self, data_dict, X, y=None, params=None, logger=None):
        """XGBoost의 fit 파라미터 구성"""
        from .._data_wrapper import unwrap

        fit_params = {}
        if params is not None and params.get('verbosity', 0) > 0:
            fit_params['verbose'] = True
        else:
            fit_params['verbose'] = False

        # data_dict에서 데이터 추출
        train_X, train_v_X = data_dict['X']
        if 'y' in data_dict:
            train_y, train_v_y = data_dict['y']
        else:
            train_y, train_v_y = None, None

        # eval_set 구성
        if self.eval_mode and self.eval_mode != 'none' and train_v_X is not None and train_v_y is not None:
            train_X_native = unwrap(train_X)
            train_y_native = unwrap(train_y)
            train_v_X_native = unwrap(train_v_X)
            train_v_y_native = unwrap(train_v_y)

            if self.eval_mode == 'valid':
                fit_params['eval_set'] = [(train_v_X_native, train_v_y_native)]
            elif self.eval_mode == 'both':
                fit_params['eval_set'] = [(train_X_native, train_y_native), (train_v_X_native, train_v_y_native)]

        return fit_params

    def _get_feature_importances(processor, importance_type):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.n_features_in_))
        booster = obj.get_booster()
        scores = booster.get_score(importance_type=importance_type)

        return pd.Series(
            [scores.get(f, 0) for f in input_vars],
            index=input_vars, name = 'importance'
        )

    def _get_evals_result(processor):
        obj = processor.obj
        evals_result =  obj.evals_result() if hasattr(obj, 'evals_result') else {}
        return pd.concat(
            [pd.DataFrame(v).stack().rename(k) for k, v in evals_result.items()], axis=1
        ).stack()

    def _get_trees(processor):
        obj = processor.obj
        booster = obj.get_booster()
        return booster.trees_to_dataframe()


XGBoostAdapter.result_objs = {
    'feature_importances': (XGBoostAdapter._get_feature_importances, True),
    'evals_result': (XGBoostAdapter._get_evals_result, True), 
    'trees': (XGBoostAdapter._get_trees, False)
}


"""
Default adapter for sklearn-compatible models
"""

from ._base import ModelAdapter


class DefaultAdapter(ModelAdapter):
    """Default adapter for models that don't support eval_set

    일반적인 sklearn 모델들을 위한 기본 어댑터
    eval_set을 지원하지 않으므로 일반 fit()만 수행
    """

    def get_fit_params(self, data_dict, X, y=None, params=None, logger=None):
        """기본 fit 파라미터 (추가 파라미터 없음)

        eval_set 관련 파라미터는 무시합니다.
        """
        # eval_set을 지원하지 않으므로 빈 dict 반환
        return {}
